#%% Function to insert sub arrays into a larger array
def insert_subarray(A_large, A_small, indices):
    '''
    # A_large: pre-allocated main sympy array to insert a sympy sub-array
    # A_small: A sympy sub-array to fill into A_large
    # indices: Indices where A_small shall be filled into A_large ([rows],[cols])
    '''

    row_idx, col_idx = indices

    # Column vector
    if A_small.rows != 1 and A_small.cols == 1:
        for i, r in enumerate(row_idx):
            A_large[r, col_idx[0]] += A_small[i, 0]
    
    # row vector
    elif A_small.rows == 1 and A_small.cols != 1:
        for i, r in enumerate(col_idx): 
            A_large[row_idx[0], r] += A_small[0, i]

    # Full array
    else:  
        for i, r in enumerate(row_idx):
            for j, c in enumerate(col_idx):
                A_large[r, c] += A_small[i, j]

#%% Function to sort letters and numbers within a list
def sort_key(sym):
    import re

    match = re.match(r"([a-zA-Z]+)(\d+)", sym.name)  # Extract prefix and number
    if match:
        prefix, num = match.groups()
        return (prefix, int(num))  # Sort alphabetically, then numerically
    return (sym.name, 0)  # Default case (if no match)

#%% Function to convert sympy expressions to a numerical framework (Numpy, Jax, etc.)
def function_save(FileName, functions_to_save, module):
    '''
    # FileName:          Name of the generated .py file
    # functions_to_save: A Python Dictonary of varibale name and the lambdified function "{var: lambdified_func}"
    # module:            Set the numerical framework "numpy", "jax" etc.
    '''

    import inspect
    import ast
    import importlib

    # Importing module (Switch to include e.g., jax or numpy)
    if module == "jax":
        module = "jax.numpy"
    # Import the module dynamically
    try:
        mod = importlib.import_module(module)
    except ImportError:
        print(f"Error: Module '{module}' could not be imported.")
        return

    # Write module import statement at the beginning of the file
    with open(FileName, "w") as file:
        file.write(f"import {module}\n\n")

    # Extracting source code adding prefix (e.g., numpy.cos or jax.numpy.sin)
    functions = {}

    # Looping though each function
    for name, func in functions_to_save.items():
        Module_attribute = set()

        # Obtaining the source code along with it's abtract syntax tree (ast)
        source_code = inspect.getsource(func) 
        tree = ast.parse(source_code) 
        input_params = list(inspect.signature(func).parameters.keys())

        # Walk through the AST to detect module attributes
        for node in ast.walk(tree):
            # Check if instance contain a "Name" e.g., "Name(id='x')" or "Name(id='cos')" AND check if "id" within said "Name" exists in the module
            if isinstance(node, ast.Name) and hasattr(node, "id") and node.id in dir(mod):
                if node.id in input_params:  # Avoid prefixing input parameters
                    print(f"Warning! Input parameter '{node.id}' conflicts with '{module}'. Ignoring prefix.")
                elif f"{module}.{node.id}" not in source_code:  # Avoid duplicate prefixing
                    Module_attribute.add(node.id)
                    source_code = source_code.replace(node.id, f"{module}.{node.id}")

        # Store modified function source and rename it
        functions[name] = source_code.replace('_lambdifygenerated', name)

        # Append modified function to the file
        with open(FileName, "a") as file:
            file.write(functions[name] + "\n")

        # Print detected attributes and final function source for debugging
        print('-' * 90)
        print(f"Detected attributes in function '{name}': {Module_attribute}\n")
        print(functions[name])


import json
import numpy as np

#%% packing of input parameters
def pack(model,name):

    # Function to convert np.array to list
    def convert_np_arrays(obj):
        if isinstance(obj, dict):
            return {k: convert_np_arrays(v) for k, v in obj.items()}
        elif isinstance(obj, list):
            return [convert_np_arrays(i) for i in obj]
        elif isinstance(obj, np.ndarray):
            return obj.tolist()
        else:
            return obj

    # Convert np.array objects to lists
    model_serialized = convert_np_arrays(model)

    # Save the cantilever_beam_model to a JSON file
    with open(name + '.json', 'w') as fid:
        json.dump(model_serialized, fid)

#%% unpacking of input parameters
def unpack(json_file):

    # Load the cantilever_beam_model from the JSON file
    with open(json_file + '.json', 'r') as fid:
        model = json.load(fid)

    # Convert lists to numpy arrays in the nested dictionaries
    for pars in model.keys():
        if isinstance(model[pars], list) and len(model[pars]) > 0 and isinstance(model[pars][0], dict):
            # Handle list of dictionaries
            for item in model[pars]:
                for key, value in item.items():
                    if isinstance(value, list):
                        item[key] = np.array(value)
        else: # isinstance(model[pars], dict):
            # Handle dictionary
            for key, value in model[pars].items():
                if isinstance(value, list):
                    model[pars][key] = np.array(value)
    
    return model


import numpy as np
import scipy as sp
import jax
import jax.numpy as jnp
import matplotlib.pyplot as plt
import matplotlib.patches as patches
import typing
from scipy.sparse import coo_array
from yafem.nodes import nodes

#%% element class
class core_elem:

    #%% class constructor
    def __init__(self, my_nodes,pars):

        self.linear_M = True
        self.linear_K = False
        self.linear_C = False

         # link the nodes to the element
        self.my_nodes = my_nodes       
        
        # extract parameters and assign default values
        self.extract_pars_core(pars)

    def extract_pars_core(self,pars):
        
        # this is stored for packing/unpacking
        self.my_pars = pars

        # set the default element tag
        self.tag = pars.get("tag",0)

        # extract nodal labels
        self.nodal_labels = pars.get("nodal_labels", np.array([1]))
        
        # extract nodal coordinates
        self.nodal_coords = self.my_nodes.find_coords(self.nodal_labels)
    
        # temperature controlled dofs
        self.dofs_q = np.array(pars.get("dofs_q", []), dtype=np.int32).reshape(-1, 2) if "dofs_q" in pars else np.zeros((0, 2), dtype=np.int32)

    #%% compute the mass matrix
    def compute_M(self,u=None,v=None,q=None,t=None,i=None):
        return self.M
    
    #%% compute the damping matrix
    def compute_C(self,u,v,q=None,t=None,i=None):
        return self.C

    #%% compute the stiffness matrix
    def compute_K(self,u,v,q=None,t=None,i=None):
        return self.K
    
     #%% compute the internal force
    def compute_f(self,u,v,a,q=None,t=None,i=None):

        # compute the restoring force
        f = self.K @ u + self.C @ v + self.M @ a

        if hasattr(self, 'B'):

            # compute the strain/cross-sectional deformation
            self.e = self.B @ u

            # compute the stress/cross-sectional force
            self.s = self.D @ self.e

        # store the state variables
        self.u = u
        self.v = v
        self.a = a
        self.f = f
        self.t = t 
        self.i = i

        # return the internal force
        return f

    #%% compute the restoring force
    def compute_r(self,u,v,q=None,t=None,i=None):

        # compute the restoring force
        r = self.K @ u + self.C @ v

        if hasattr(self, 'B'):

            # compute the strain/cross-sectional deformation
            self.e = self.B @ u

            # compute the stress/cross-sectional force
            self.s = self.D @ self.e

        # store the state variables
        self.u = u
        self.v = v
        self.r = r
        self.t = t 
        self.i = i

        # return the restoring force
        return r

    #%% Function for compute the collocation matrix Zu and Zq
    @jax.jit
    def __compute_collocation(dof_e, dofs):
        def indices(dof_e, dofs):
            match = jnp.all(dofs == dof_e, axis=1)
            return jnp.argmax(match), jnp.any(match)

        idx, found = jax.vmap(indices, in_axes=(0, None))(dof_e, dofs)
        rows = jnp.where(found, jnp.arange(jnp.shape(found)[0]), 0)
        cols = jnp.take(idx, rows)
        data = jnp.ones_like(rows, dtype=bool)

        return rows, cols, data, found, found.all()

    def __compute_collocation_matrix(self, dof_e, dofs):
        if jnp.shape(dofs)[0] != 0:
            rows, cols, data, found, all_true = core_elem.__compute_collocation(dof_e, dofs)

            if not all_true:
                rows = rows[found]
                cols = cols[found]
                data = np.ones_like(rows, dtype=bool)
        else:
            rows, cols, data = [], [], []

        return coo_array((data, (rows, cols)), shape=(dof_e.shape[0], dofs.shape[0]), dtype=bool).tocsr()

    def compute_Zu(self, dofs):
        self.Zu = self.__compute_collocation_matrix(self.dofs, dofs)

    def compute_Zq(self, dofs_q):
        self.Zq = self.__compute_collocation_matrix(self.dofs_q, dofs_q)

    #%% reset the element state
    def reset(self):

        self.u = np.zeros(self.dofs.shape[0], dtype=int)
        self.v = self.u.copy()
        self.a = self.u.copy()
        self.q = np.zeros(self.dofs_q.shape[0], dtype=int)

    #%% plot the element
    def plot(self,ax,x=None, y=None, z=None, color=None):
        pass
    
    #%% compute element results (e.g., strain and stress)
    def compute_results(self):
        pass

    #%% save element results in paraview
    def dump_to_paraview(self):
        pass



import numpy as np
import scipy as sp
from scipy.linalg import block_diag
from yafem.nodes import nodes
from yafem.elem.core_elem import core_elem
from yafem.elem.tri6_func import *
from mpl_toolkits.mplot3d.art3d import Line3DCollection

#%% element_MCK class
class tri6(core_elem):

#%% class constructor
    def __init__(self, my_nodes, pars):

        # superclass constructor
        super().__init__(my_nodes,pars)

        self.linear_M = True
        self.linear_K = True
        self.linear_C = True

        # link the nodes to the element
        self.my_nodes = my_nodes

        # extract parameters and assign default values
        self.extract_pars(pars)

        # element dofs
        self.element_dofs(6)

        # vector to corners node 1 to 2 and node 1 to 3
        nodal_corners = self.nodal_coords
        
        # node 2 - node 1
        xp = nodal_corners[1,:] - nodal_corners[0,:]
        
        # normal vector to the plane
        zp = np.linalg.cross(xp,nodal_corners[2,:] - nodal_corners[0,:])

        xp = xp / np.linalg.norm(xp)
        zp = zp / np.linalg.norm(zp)

        # unit vector in the y-direction
        yp = np.cross(zp, xp)

        pc = np.mean(self.nodal_coords,axis=0)

        # Local axis xp, yp and zp shall be column vector
        self.T = np.array([xp, yp, zp])

        self.xp = xp
        self.yp = yp
        self.zp = zp

        xe = (self.nodal_coords - pc) @ np.array([xp, yp]).T
        self.xe = xe

        # Transformation matrix for the displacement vector of a single node
        # This is a triangle element, look into Filippa's book for further theoretical basis
        self.G = np.kron(np.eye(12), self.T)
        self.G = np.delete(self.G, np.arange(5,36,6), axis=0)

        self.gp_xy = gp_xy_6()
        self.gw_xy = gw_xy_6().squeeze()
        self.gp_z  = gp_z().squeeze()
        self.gw_z  = gw_z().squeeze()

        D_val = D_func(self.E,self.nu)
        I_val = I()
        h_half = self.h * 0.5

        self.Kl = np.zeros([30,30])
        self.Ml = np.zeros([30,30])

        xe_flattend = xe.flatten()

        # volume of the element
        self.area = A_func(xe_flattend)
        self.vol = self.area * self.h

        gp_plot = []

        # loop over the gauss points (in-plane)
        for i, gp_xyi in enumerate(self.gp_xy):

            # jacobian matrix
            J_val = jac(xe_flattend, gp_xyi)
            det_J_val = np.linalg.det(J_val)/2
                
            # inverse of jacobian matrix
            P_val = np.linalg.solve(J_val, I_val)
            P_val_block = block_diag(P_val, P_val, P_val).T

            gp_plot.append(self.nodal_coords.T @ phi(gp_xyi))

            B2 = B_z(xe_flattend, gp_xyi)

            # loop over the lobatto points (thickness)
            for j, gp_zj in enumerate(self.gp_z):
                  
                # half the height of the element multiplied with the gauss point
                zj = gp_zj * h_half
                  
                # displacement interpolation matrix
                N_val = N_func(xe_flattend, gp_xyi, zj)

                # strain interpolation matrix                  
                B1 = P_val_block @ B_xy(xe_flattend, gp_xyi, zj)
                B_val = B_b() @ np.vstack((B1, B2))

                # Gauss and Lobatto weights
                weight = self.gw_xy[i] * self.gw_z[j] * det_J_val * h_half

                # stiffness and mass matrix
                self.Kl += B_val.T @ D_val @ B_val * weight
                self.Ml += N_val.T @ N_val * self.rho * weight

        self.K = self.G.T @ self.Kl @ self.G
        self.M = self.G.T @ self.Ml @ self.G

        self.gp_plot = np.hstack(gp_plot).T

#%% extract parameters and assign default values
    def extract_pars(self,pars):

        # this is the element class used in packing/unpacking
        self.my_pars['elem'] = 'tri6'

        self.E   = pars.get('E', 210e9) # young's modulus
        self.nu  = pars.get('nu', 0.3) # poisson's ratio
        self.rho = pars.get('rho', 7850) # material density
        self.h   = pars.get('h', 5e-3) # element thickness
        self.k   = pars.get('k', 5/6)  # shear factor
        self.nodal_labels = pars.get("nodal_labels", [1, 2, 3, 4, 5, 6])
        self.nodal_coords = self.my_nodes.find_coords(self.nodal_labels) # extract nodal coordinates
        self.type    = pars.get("type", "ps")  # type of analysis (ps = plane stress, pe = plane strain, ax = axisymmetric)

        # temperature controlled dofs
        self.dofs_q = pars.get('dofs_q', np.zeros((0, 2), dtype=np.int32)).astype(np.int32)

#%% element dofs
    def element_dofs(self, dofs_per_node):
        self.dofs = np.empty([dofs_per_node*6,2],dtype=int)

        self.dofs[0:dofs_per_node,0]                 = self.nodal_labels[0] # Label of first node
        self.dofs[dofs_per_node*1:dofs_per_node*2,0] = self.nodal_labels[1] # Label of second node
        self.dofs[dofs_per_node*2:dofs_per_node*3,0] = self.nodal_labels[2] # Label of third node
        self.dofs[dofs_per_node*3:dofs_per_node*4,0] = self.nodal_labels[3] # Label of fourth node
        self.dofs[dofs_per_node*4:dofs_per_node*5,0] = self.nodal_labels[4] # Label of fifth node
        self.dofs[dofs_per_node*5:dofs_per_node*6,0] = self.nodal_labels[5] # Label of sixth node
        self.dofs[:,1] = np.tile(np.arange(0,dofs_per_node), 6) + 1 # Dofs of all nodes
    
        return self.dofs

#%% Plot 3d elements
    def plot(self, ax, x=None, y=None, z=None, color='cyan'):
        if x is None: x = self.nodal_coords[:, 0]
        if y is None: y = self.nodal_coords[:, 1]
        if z is None: z = self.nodal_coords[:, 2]
            
        # Subdivide 6-node triangle into 4 smaller 3-node triangles
        triangles = np.array([[0, 3, 5],    # Node 1 4 6
                              [3, 1, 4],    # Node 4 2 5
                              [5, 4, 2],    # Node 6 5 3
                              [3, 4, 5]])   # Node 4 5 6

         # Collect surface triangles using a lambda function
        surfaces_func = lambda row: \
            [[x[triangles[row, 0]], y[triangles[row, 0]], z[triangles[row, 0]]],
             [x[triangles[row, 1]], y[triangles[row, 1]], z[triangles[row, 1]]],
             [x[triangles[row, 2]], y[triangles[row, 2]], z[triangles[row, 2]]]]

        surfaces = [surfaces_func(0),
                    surfaces_func(1),
                    surfaces_func(2),
                    surfaces_func(3)]

        # Convert edges to line segments in the correct format
        lines = [[[x[0], y[0], z[0]], [x[3], y[3], z[3]]], # [0, 3]
                 [[x[3], y[3], z[3]], [x[1], y[1], z[1]]], # [3, 1]
                 [[x[1], y[1], z[1]], [x[4], y[4], z[4]]], # [1, 4]
                 [[x[4], y[4], z[4]], [x[2], y[2], z[2]]], # [4, 2]
                 [[x[2], y[2], z[2]], [x[5], y[5], z[5]]], # [2, 5]
                 [[x[5], y[5], z[5]], [x[0], y[0], z[0]]]] # [5, 0]

        return lines, surfaces


import numpy as np
import matplotlib.pyplot as plt
import matplotlib.patches as patches
from scipy.linalg import block_diag
from yafem.nodes import nodes
from yafem.elem.core_elem import core_elem
from yafem.elem.snapback_func import *

class snapback(core_elem):

    #%% class constructor
    def __init__(self, my_nodes, pars):

        # superclass constructor
        super().__init__(my_nodes,pars)

        # link the nodes to the element
        self.my_nodes = my_nodes
        
        # extract parameters and assign default values
        self.extract_pars(pars)

        self.element_dofs(1)

        # initialization of matrices
        self.M = np.zeros([2,2])
        self.C = self.M .copy()
        self.K = snapback_K(self.W,self.A,self.E,self.L,self.a,[0,0])

    #%% extract parameters
    def extract_pars(self, pars):
        self.E            = pars.get("E", 1.0) # young's modulus
        self.L            = pars.get("L", 1.0) # element length
        self.A            = pars.get("A", 1.0) # element cross section area
        self.W            = pars.get("W", 0.3) # stiffness ratio
        self.a            = pars.get("a", self.L * np.sin(np.pi/3)) # displacement step
        self.nodal_labels = pars.get("nodal_labels", [1, 2])

        # temperature controlled dofs
        self.dofs_q = np.array(pars.get("dofs_q", []), dtype=np.int32).reshape(-1, 2) if "dofs_q" in pars else np.zeros((0, 2), dtype=np.int32)
    
    #%% Computing element dofs
    def element_dofs(self, dofs_per_node):

        self.dofs = np.empty([dofs_per_node*2,2],dtype=int)

        self.dofs[0:dofs_per_node,0] = self.nodal_labels[0] # Label of first node
        self.dofs[dofs_per_node:,0]  = self.nodal_labels[1] # Label of second node
        self.dofs[:,1] = np.tile(np.arange(0,dofs_per_node), 2) + 1 # Dofs of both nodes
    
        return self.dofs

    #%% Restoring force
    def compute_r(self,u,v,gq,t,i):

        # Link to the restoring force vector
        r = snapback_r(self.W,self.A,self.E,self.L,self.a,u).squeeze()

        return r

    #%% Tangent stiffness
    def compute_K(self,u,v,gq,t,i):

        # Link to the restoring force vector
        K = snapback_K(self.W,self.A,self.E,self.L,self.a,u)

        return K



import numpy as np
import matplotlib.pyplot as plt
import matplotlib.patches as patches
from scipy.linalg import block_diag
from yafem.nodes import nodes
from yafem.elem.core_elem import core_elem
from yafem.elem.beam3d_circ_func import *
from yafem.elem.beam3d_gen_func import *

class beam3d(core_elem):
    #%% class constructor
    def __init__(self, my_nodes, pars):

        # superclass constructor
        super().__init__(my_nodes,pars)

        self.linear_M = True
        self.linear_K = True
        self.linear_C = True

        # extract parameters and assign default values
        self.extract_pars(pars)

        # set the number of dofs per node
        self.element_dofs(6)

        # Coordinate transformation matrix calculation
        r = self.nodal_coords[1] - self.nodal_coords[0]
        r /= np.linalg.norm(r)
        
        # Determine a perpendicular vector s and orthogonalize
        s = np.array([0.0, 1.0, 0.0]) if r[2] != 0 else np.array([-r[1], r[0], 0.0])
        s -= np.dot(s, r) * r
        s /= np.linalg.norm(s)

        # Compute t as the orthogonal cross product
        t = np.cross(r, s)
        t /= np.linalg.norm(t)

        # Create transformation matrix
        self.T = np.array([r, s, t])
        
        self.G = np.zeros((12, 12))
        self.G[0:3, 0:3] = self.T
        self.G[3:6, 3:6] = self.T
        self.G[6:9, 6:9] = self.T
        self.G[9:12, 9:12] = self.T

        if self.shape == 'circular':

            # Variables
            variables = [self.D1, 
                         self.D2, 
                         self.E, 
                         self.G1, 
                         self.H, 
                         self.L, 
                         self.alpha, 
                         self.fa, 
                         self.fby, 
                         self.fbz, 
                         self.k0a, 
                         self.k0b, 
                         self.rho, 
                         self.theta]
            
            self.Kl  = beam3d_circ_Kl(*variables)
            self.Ml  = beam3d_circ_Ml(*variables)
            self.rl  = beam3d_circ_rl(*variables)
            self.Nl  = beam3d_circ_Nl(*variables)
            self.Bl  = beam3d_circ_Bl(*variables)

            # cross-section stiffness matrix
            self.D = beam3d_circ_Dcs_mid(*variables)

        else: # self.shape == 'generic':
            
            # Variables
            variables = [self.A, 
                         self.E, 
                         self.G1, 
                         self.Ixx, 
                         self.Iyy, 
                         self.Jv, 
                         self.L, 
                         self.alpha, 
                         self.fa, 
                         self.fby, 
                         self.fbz, 
                         self.k0a, 
                         self.k0b, 
                         self.rho, 
                         self.theta]
            
            self.Kl  = beam3d_gen_Kl(*variables)
            self.Ml  = beam3d_gen_Ml(*variables)
            self.rl  = beam3d_gen_rl(*variables)
            self.Nl  = beam3d_gen_Nl(*variables)
            self.Bl  = beam3d_gen_Bl(*variables)

            # cross-section stiffness matrix
            self.D = beam3d_gen_Dcs_mid(*variables)

        #%% Global reference

        # Stiffness matrix in global coordinate system
        self.K = self.G.T @ self.Kl @ self.G
        self.M = self.G.T @ self.Ml @ self.G

        # damping matrix in global coordinates
        self.C = np.zeros_like(self.K)

        # strain interpolation matrix in global coordinates
        self.B = self.Bl @ self.G

        # displacement interpolation matrix in global coordinates
        self.N = self.Nl @ self.G

        # local to global coordinate transformation
        self.r = self.G.T @ self.rl     
 
    #%% extract parameters
    def extract_pars(self, pars):
        
        # this is the element class used in packing/unpacking
        self.my_pars['elem'] = 'beam3d'

        self.shape    = pars.get("shape", 'circular') # Youngs modulus

        if self.shape == 'circular':
            self.D1   = pars.get("D1", 20.0) # Outer diameter of pipe in firste node
            self.D2   = pars.get("D2", 20.0) # Outer diameter of pipe in second node
            self.H    = pars.get("H", 2) # thickness of pipe in both ends

            # Cross section area and inertia as functions of element diameter
            self.A1 = np.pi * (self.D1/2)**2 - np.pi * ((self.D1/2)-self.H)**2
            self.A2 = np.pi * (self.D2/2)**2 - np.pi * ((self.D2/2)-self.H)**2
            self.I1 = np.pi/4 * (self.D1/2)**4 - np.pi/4 * ((self.D1/2)-self.H)**4
            self.I2 = np.pi/4 * (self.D2/2)**4 - np.pi/4 * ((self.D2/2)-self.H)**4
            self.J1 = 2*self.I1  # Polar moment of inertia
            self.J2 = 2*self.I2  # Polar moment of inertia
        
        else: # self.shape == 'generic':
            self.A    = pars.get("A", 200.0) # Cross-sectional areal
            self.Ixx  = pars.get("Ixx", 1000.0) # second area moment about xx
            self.Iyy  = pars.get("Iyy", 1000.0) # second area moment about yy
            self.Jv   = pars.get("Jv", self.Ixx + self.Iyy) # Torsional constant
 
        self.E     = pars.get("E", 210e3) # Youngs modulus
        self.G1    = pars.get("G", 81e3) # Shear modulus
        self.k0a   = pars.get("k0a", 0.0) # Axial Winkler stiffness
        self.k0b   = pars.get("k0b", 0.0) # Lateral Winkler stiffness
        self.rho   = pars.get("rho", 7850/1e9) # Density
        self.fa    = pars.get("fa", 0.0) # Axial element destributed force
        self.fby   = pars.get("fby", 0.0) # Lateral element destributed force in y-direction
        self.fbz   = pars.get("fbz", 0.0) # Lateral element destributed force in z-direction
        self.alpha = pars.get("alpha", 0.0) # coefficient of thermal expansion 
        self.theta = pars.get("theta", 0.0) # Thermal loading
        self.nodal_labels = pars.get("nodal_labels", [1, 2])
        
        # extract nodal coordinates
        self.nodal_coords = self.my_nodes.find_coords(self.nodal_labels)
        self.L = np.linalg.norm(self.nodal_coords[1] - self.nodal_coords[0])
    
        # temperature controlled dofs
        self.dofs_q = np.array(pars.get("dofs_q", []), dtype=np.int32).reshape(-1, 2) if "dofs_q" in pars else np.zeros((0, 2), dtype=np.int32)

    #    self.nu    = pars.get("nu", 0.3)
    #    self.J     = pars.get("J", 1.0)
    
    #%% Computing element dofs
    def element_dofs(self, dofs_per_node):

        self.dofs = np.empty([dofs_per_node*2,2],dtype=int)

        self.dofs[0:dofs_per_node,0] = self.nodal_labels[0] # Label of first node
        self.dofs[dofs_per_node:,0]  = self.nodal_labels[1] # Label of second node
        self.dofs[:,1] = np.tile(np.arange(0,dofs_per_node), 2) + 1 # Dofs of both nodes
    
        return self.dofs

    #%% Plot 3d elements
    def plot(self, ax, x=None, y=None, z=None, color='k-'):
        if x is None: x = self.nodal_coords[:, 0]
        if y is None: y = self.nodal_coords[:, 1]
        if z is None: z = self.nodal_coords[:, 2]

        # Collect lines
        lines = [[[x[0], y[0], z[0]],[x[1], y[1], z[1]]]]

        return lines, None
    
    def dump_to_paraview(self):
        # here it goes the dump_to_paraview implementation for the beam3d element
        pass



# def function_save(FileName,functions_to_save,module):
#     import inspect
#     import ast

#     #%% Importing module (Switch to include e.g., jax or numpy)
#     if module == 'jax':
#         exec('import ' + module + '.numpy') 
#         module = module + '.numpy'
#     else:
#         exec('import ' + module) 

#     #%% Importing module
#     with open(FileName, "w") as file:
#         file.write('import ' + module + " \n\n")

#     #%% Extracting source code adding prefix (e.g., numpy.cos or jax.numpy.sin)
#     functions = {}
    
#     # Looping though each function
#     for name, func in functions_to_save.items():
#         Module_attribute = set()

#         # Obtaining the source code along with it's abtract syntax tree (ast)
#         source_code = inspect.getsource(func) 
#         tree = ast.parse(source_code) 

#         # Input parameters
#         input_params = list(inspect.signature(func).parameters.keys()) 

#         # Looping through the ast
#         for node in ast.walk(tree):
            
#             # Check if instance contain a "Name" e.g., "Name(id='x')" or "Name(id='cos')" AND check if "id" within said "Name" exists in the module
#             if isinstance(node, ast.Name) and node.id in dir(eval(module)): 
#                 if node.id in input_params: # Warning if input parameter(s) clash with detected attribute
#                     print('Warning! Input parameter {' + node.id +  '} share attribute with {' + module + '}. Ignoring prefix for {' + node.id + '}.')
#                 else:
#                     if module + '.' + node.id in source_code: # Case with repeated attributes
#                         continue
#                     else: 
#                         Module_attribute.add(node.id) # set of detected attributes
#                         source_code = source_code.replace(node.id, f'{module}.{node.id}') # Applying prefix to detected attributes

#         # Adding modified source_code with prefix and renaming the function
#         functions[name] = source_code
#         functions[name] = functions[name].replace('_lambdifygenerated', name)

#         #%% Adding function the file
#         with open(FileName, "a") as file:
#             file.write(functions[name] + '\n')

#         #%% Printing attribute(s) and function(s) for quick validation
#         print('------------------------------------------------------------------------------------------')
#         print('Detected attributes: ' + str(Module_attribute))
#         print(' ')
#         print(functions[name])


def function_save(FileName, functions_to_save, module):
    import inspect
    import ast
    import importlib

    #%% Importing module (Switch to include e.g., jax or numpy)
    if module == "jax":
        module = "jax.numpy"
    # Import the module dynamically
    try:
        mod = importlib.import_module(module)
    except ImportError:
        print(f"Error: Module '{module}' could not be imported.")
        return

    # Write module import statement at the beginning of the file
    with open(FileName, "w") as file:
        file.write(f"import {module}\n\n")

    #%% Extracting source code adding prefix (e.g., numpy.cos or jax.numpy.sin)
    functions = {}

    # Looping though each function
    for name, func in functions_to_save.items():
        Module_attribute = set()

        # Obtaining the source code along with it's abtract syntax tree (ast)
        source_code = inspect.getsource(func) 
        tree = ast.parse(source_code) 
        input_params = list(inspect.signature(func).parameters.keys())

        # Walk through the AST to detect module attributes
        for node in ast.walk(tree):
            # Check if instance contain a "Name" e.g., "Name(id='x')" or "Name(id='cos')" AND check if "id" within said "Name" exists in the module
            if isinstance(node, ast.Name) and hasattr(node, "id") and node.id in dir(mod):
                if node.id in input_params:  # Avoid prefixing input parameters
                    print(f"Warning! Input parameter '{node.id}' conflicts with '{module}'. Ignoring prefix.")
                elif f"{module}.{node.id}" not in source_code:  # Avoid duplicate prefixing
                    Module_attribute.add(node.id)
                    source_code = source_code.replace(node.id, f"{module}.{node.id}")

        # Store modified function source and rename it
        functions[name] = source_code.replace('_lambdifygenerated', name)

        # Append modified function to the file
        with open(FileName, "a") as file:
            file.write(functions[name] + "\n")

        # Print detected attributes and final function source for debugging
        print('-' * 90)
        print(f"Detected attributes in function '{name}': {Module_attribute}\n")
        print(functions[name])



from yafem.elem.MCK import MCK
from yafem.elem.beam2d import beam2d
from yafem.elem.beam2d_cr import beam2d_cr
from yafem.elem.beam3d import beam3d
from yafem.elem.shell4 import shell4
from yafem.elem.tri6 import tri6
from yafem.elem.tet10 import tet10
from yafem.elem.snapback import snapback
from yafem.elem.snapthrough import snapthrough

__all__ = ['MCK',
           'beam2d',
           'beam2d_cr',
           'beam3d',
           'shell4',
           'tri6',
           'tet10',
           'snapback',
           'snapthrough',
           ]



import numpy as np
import scipy as sp
from scipy.sparse import coo_array, csr_array
from yafem.nodes import nodes
from scipy.linalg import block_diag
from yafem.elem.core_elem import core_elem
from yafem.elem.tet10_func import *
from mpl_toolkits.mplot3d.art3d import Line3DCollection

#%% element_MCK class
class tet10(core_elem):

#%% class constructor
    def __init__(self, my_nodes, pars):

        # superclass constructor
        super().__init__(my_nodes,pars)

        self.linear_M = True
        self.linear_K = True
        self.linear_C = True

        # extract parameters and assign default values
        self.extract_pars(pars)

        # element dofs
        self.element_dofs(3)

        D_mat = D(self.E,self.nu)

        self.gp = gp()
        self.gw = gw()

        n_coord_flattend = self.nodal_coords.flatten()

        B1 = np.array([[1,0,0 , 0,0,0 , 0,0,0],
                       [0,0,0 , 0,1,0 , 0,0,0],
                       [0,0,0 , 0,0,0 , 0,0,1],
                       [0,0,0 , 0,0,1 , 0,1,0],
                       [0,0,1 , 0,0,0 , 1,0,0],
                       [0,1,0 , 1,0,0 , 0,0,0]], dtype=bool)
        
        I = np.array([[0, 0, 0], 
                      [1, 0, 0], 
                      [0, 1, 0], 
                      [0, 0, 1]], dtype=bool)

        # allocating K and M
        self.K = np.zeros([30,30])
        self.M = np.zeros([30,30])

        self.vol = V(*n_coord_flattend)
        jac3 = np.zeros((12, 9))

        for i, gpi in enumerate(self.gp):     

                # displacement interpolation matrix
                N_mat = N(n_coord_flattend, gpi)

                # jacobian matrix
                P_val = np.linalg.solve(jac(n_coord_flattend, gpi),I)
                
                # block diagonal
                jac3[0:4, 0:3]  = P_val
                jac3[4:8, 3:6]  = P_val
                jac3[8:12, 6:9] = P_val

                # strain interpolation matrix
                B_mat = B1 @ jac3.T @ B2(n_coord_flattend, gpi)

                # Gauss weights
                weight = self.gw[i]
                    
                # stiffness and mass matrix
                self.K += B_mat.T @ D_mat @ B_mat * self.vol * weight
                self.M += N_mat.T @ N_mat * self.rho * self.vol * weight

#%% extract parameters and assign default values
    def extract_pars(self,pars):

        # this is the element class used in packing/unpacking
        self.my_pars['elem'] = 'tet10'

        self.E   = pars.get('E', 210e9) # young's modulus
        self.nu  = pars.get('nu', 0.3) # poisson's ratio
        self.rho = pars.get('rho', 7850) # material density
        self.h   = pars.get('h', 5e-3) # element thickness
        self.k   = pars.get('k', 5/6)  # shear factor
        self.nodal_labels = pars.get("nodal_labels", [1, 2, 3, 4, 5, 6, 7, 8, 9, 10])
        self.nodal_coords = self.my_nodes.find_coords(self.nodal_labels) # extract nodal coordinates
        self.gauss_order  = pars.get("gauss_order", 2)  # number of Gauss points
        self.type    = pars.get("type", "ps")  # type of analysis (ps = plane stress, pe = plane strain, ax = axisymmetric)

        # temperature controlled dofs
        self.dofs_q = pars.get('dofs_q', np.zeros((0, 2), dtype=np.int32)).astype(np.int32)

#%% element dofs
    def element_dofs(self, dofs_per_node):
        self.dofs = np.empty([dofs_per_node*10,2],dtype=int)

        self.dofs[0:dofs_per_node,0]                  = self.nodal_labels[0] # Label of first node
        self.dofs[dofs_per_node*1:dofs_per_node*2,0]  = self.nodal_labels[1] # Label of second node
        self.dofs[dofs_per_node*2:dofs_per_node*3,0]  = self.nodal_labels[2] # Label of third node
        self.dofs[dofs_per_node*3:dofs_per_node*4,0]  = self.nodal_labels[3] # Label of fourth node
        self.dofs[dofs_per_node*4:dofs_per_node*5,0]  = self.nodal_labels[4] # Label of fifth node
        self.dofs[dofs_per_node*5:dofs_per_node*6,0]  = self.nodal_labels[5] # Label of sixth node
        self.dofs[dofs_per_node*6:dofs_per_node*7,0]  = self.nodal_labels[6] # Label of sevetnth node
        self.dofs[dofs_per_node*7:dofs_per_node*8,0]  = self.nodal_labels[7] # Label of eigthth node
        self.dofs[dofs_per_node*8:dofs_per_node*9,0]  = self.nodal_labels[8] # Label of ninthth node
        self.dofs[dofs_per_node*9:dofs_per_node*10,0] = self.nodal_labels[9] # Label of tenth node
        self.dofs[:,1] = np.tile(np.arange(0,dofs_per_node), 10) + 1 # Dofs of all nodes
    
        return self.dofs

#%% Plot 3d elements
    def plot(self, ax, x=None, y=None, z=None, color='cyan'):
        if x is None: x = self.nodal_coords[:, 0]
        if y is None: y = self.nodal_coords[:, 1]
        if z is None: z = self.nodal_coords[:, 2]
            
        # Subdivide 6-node triangle into 4 smaller 3-node triangles
        triangles = np.array([[0, 4, 6],    # Node 1 5 7 # xy triangle
                              [4, 1, 5],    # Node 5 2 6
                              [6, 5, 2],    # Node 7 6 3
                              [4, 5, 6],    # Node 5 6 7

                              [0, 7, 4],    # Node 1 8 5 # xz triangle
                              [4, 9, 1],    # Node 5 10 2
                              [7, 3, 9],    # Node 8 4 10
                              [7, 9, 4],    # Node 8 10 5

                              [0, 6, 7],    # Node 1 7 8  # yz triangle
                              [6, 2, 8],    # Node 7 3 9
                              [7, 8, 3],    # Node 8 9 4
                              [7, 6, 8],    # Node 8 7 9

                              [5, 1, 9],    # Node 2 6 10 # incline triangle
                              [2, 5, 8],    # Node 6 3 9
                              [8, 9, 3],    # Node 9 10 4
                              [8, 5, 9],    # Node 9 6 10
                              ]) 

         # Collect surface triangles using a lambda function
        surfaces_func = lambda row: \
            [[x[triangles[row, 0]], y[triangles[row, 0]], z[triangles[row, 0]]],
             [x[triangles[row, 1]], y[triangles[row, 1]], z[triangles[row, 1]]],
             [x[triangles[row, 2]], y[triangles[row, 2]], z[triangles[row, 2]]]]

        surfaces = [surfaces_func(0), # xy triangle
                    surfaces_func(1),
                    surfaces_func(2),
                    surfaces_func(3),

                    surfaces_func(4), # xz triangle
                    surfaces_func(5),
                    surfaces_func(6),
                    surfaces_func(7),

                    surfaces_func(8), # yz triangle
                    surfaces_func(9),
                    surfaces_func(10),
                    surfaces_func(11),

                    surfaces_func(12), # incline triangle
                    surfaces_func(13),
                    surfaces_func(14),
                    surfaces_func(15),
                    ]

        line_func = lambda N1,N2: [[x[N1], y[N1], z[N1]], [x[N2], y[N2], z[N2]]]

        # Convert edges to line segments in the correct format
        lines = [line_func(0,4), # xy triangle
                 line_func(4,1),
                 line_func(1,5),
                 line_func(5,2),
                 line_func(2,6),
                 line_func(6,0),

                 line_func(0,7), # xz triangle
                 line_func(7,3),
                 line_func(1,9),
                 line_func(9,3),

                 line_func(3,8), # xz triangle
                 line_func(8,2),
                ]

        return lines, surfaces
    


import numpy as np
import scipy as sp
import matplotlib.pyplot as plt
import matplotlib.patches as patches
import typing
from scipy.sparse.linalg import spsolve
from scipy.sparse import csr_array, csc_array
from yafem.model import *

class simulation:

    # recorder_fields : typing.List[np.ndarray[np.float64]] = []
    msg = "step {} converged in {} iterations with residual {}" # pre-loading a message for simulations

#%% object constructor
    def __init__(self,my_model,pars={}):

        # link the list of elements to the model
        self.my_model = my_model

        # extract the parameters and assign default values
        self.extract_pars(pars)

        # initialize recorder data
        self.recorder_data = {}

        # initialize recorder fields
        for recorder_tag in self.recorder_tags:
            print("created recorder field for: " + recorder_tag)
            self.recorder_data[recorder_tag] = np.zeros(self.my_model.step)

        # initialize solution variables
        self.u = np.zeros((self.my_model.dofs.shape[0],self.my_model.step)) # displacements
        self.v = self.u.copy()  # velocities
        self.a = self.u.copy()  # accelerations
        self.r = self.u.copy()  # restoring force
        self.f = self.u.copy()  # internal force
        
        # initialize solution variables (extended)
        self.u_x = np.zeros((self.my_model.dofs_x.shape[0],self.my_model.step)) # displacements
        self.v_x = self.u_x.copy() # velocities
        self.a_x = self.u_x.copy() # accelerations
        self.r_x = self.u_x.copy() # restoring force
        self.f_x = self.u_x.copy() # internal force
        
        self.l = np.zeros((self.my_model.dofs_u.shape[0],self.my_model.step)) # lagrange multipliers

        # initialize the mass matrix (constant)
        self.M = self.my_model.M

        # initialize the proportional damping matrix (constant)
        self.Cp = self.my_model.Cp

        # initialize the time axis
        self.t = np.arange(0, self.my_model.step * self.my_model.dt + self.my_model.dt, self.my_model.dt)
        

#%% extract parameters and assign default values
    def extract_pars(self,pars):

        # storage of input for packing/unpacking
        self.my_pars = pars

        self.tol_min       = float(pars.get('tol_min', 1e-6)) # minimum convergence tolerance
        self.iter_max      = int(pars.get('iter_max', 20)) # maximum number of iterations
        self.gamma         = pars.get('gamma', 0.5) # Newmark gamma
        self.beta          = pars.get('beta', 0.25) # Newmark beta
        self.delta         = pars.get('delta', 1.0) # Arclen parameter
        self.recorder_tags = pars.get('recorder_tags', []) # recorders

#%% dynamic analysis 2
    # time history analysis using the Newmark method (only force and temperature controlled dofs)
    def dynamic_analysis2(self, step1=0, step2=None,output=True):

        if step2 is None:
            step2 = self.my_model.step  # Default last analysis step

        # time integration loop
        for i in range(step1,step2-1):

            # perform dynamic analysis step
            self.u[:,i+1],self.v[:,i+1],self.a[:,i+1],self.f[:,i+1] = self.dynamic_analysis_step2(self.u[:,i],self.v[:,i],self.a[:,i],self.t[i],i,output)

            # record model results
            self.record_step(i+1)
            
        # return output
        return self.u,self.v,self.a,self.f
    
    # perform dynamic analysis step
    def dynamic_analysis_step2(self,u0,v0,a0,t0,i0,output):

        # Newmark predictors
        a1 = np.zeros(self.my_model.dofs.shape[0])
        v1 = v0 + self.my_model.dt * (1-self.gamma) * a0

        u1 = u0 + self.my_model.dt * v0 + self.my_model.dt**2 * (0.5-self.beta) * a0
        t1 = t0 + self.my_model.dt
        i1 = i0 + 1

        # compute the internal forces
        # r1 = self.my_model.compute_r(u1,v1,self.my_model.g_q[:,i1],t1,i1)
        f1 = self.my_model.compute_f(u1,v1,a1,self.my_model.g_q[:,i1],t1,i1)
        # print(f1)
        # residual of the equilibrium equation
        # res = self.M @ a1 + self.Cp @ v1 + r1 - self.my_model.Bf @ self.my_model.g_f[:,i1]
        res = self.Cp @ v1 + f1 - self.my_model.Bf @ self.my_model.g_f[:,i1]

        # iteration counter
        cc = 0

        while True:

            # tangent stiffness matrix
            K = self.my_model.compute_K(u1,v1,self.my_model.g_q[:,i1],t1,i1)

            # tangent damping matrix
            C = self.my_model.compute_C(u1,v1,self.my_model.g_q[:,i1],t1,i1)
            
            # Tangent mass matrix
            M = self.my_model.compute_M(u1,v1,self.my_model.g_q[:,i1],t1,i1)
            
            # jacobian of the residual           
            jac = M + (self.Cp + C) * self.gamma * self.my_model.dt + K * self.beta * self.my_model.dt**2

            # acceleration update
            da = -spsolve(jac,res)
 
            # update all kinematic quantities
            a1 = a1 + da
            v1 = v1 + da * self.gamma * self.my_model.dt
            u1 = u1 + da * self.beta  * self.my_model.dt**2

            # compute the internal forces
            # r1 = self.my_model.compute_r(u1,v1,self.my_model.g_q[:,i1],t1,i1)
            f1 = self.my_model.compute_f(u1,v1,a1,self.my_model.g_q[:,i1],t1,i1)

            # residual of the equilibrium equation
            # res = self.M @ a1 + self.Cp @ v1 + r1 - self.my_model.Bf @ self.my_model.g_f[:,i1]
            res = self.Cp @ v1 + f1 - self.my_model.Bf @ self.my_model.g_f[:,i1]

            # convergence tolerance
            tol = (np.linalg.norm(res) * self.my_model.ndof) / (self.my_model.ndof * (1 + np.linalg.norm(np.dot(self.my_model.Bf.todense(), self.my_model.g_f[:, i1]))))

            # exit condition
            if cc >= self.iter_max or tol < self.tol_min or self.my_model.linear == True:
                if output == True:
                    # print(f"step {i1} converged in {cc+1} iterations with residual {tol}")
                    print(simulation.msg.format(i1, cc + 1, tol))
                break
            else:
                cc += 1

        # return the results of the analysis step
        return u1,v1,a1,f1


#%% dynamic analysis
    # time history analysis using the Newmark method (only force and temperature controlled dofs)
    def dynamic_analysis(self, step1=0, step2=None,output=True):

        if step2 is None:
            step2 = self.my_model.step  # Default last analysis step

        # time integration loop
        for i in range(step1,step2-1):

            # perform dynamic analysis step
            self.u[:,i+1],self.v[:,i+1],self.a[:,i+1],self.r[:,i+1] = self.dynamic_analysis_step(self.u[:,i],self.v[:,i],self.a[:,i],self.t[i],i,output)

            # record model results
            self.record_step(i+1)
            
        # return output
        return self.u,self.v,self.a,self.r
    
    # perform dynamic analysis step
    def dynamic_analysis_step(self,u0,v0,a0,t0,i0,output):

        # Newmark predictors
        a1 = np.zeros(self.my_model.dofs.shape[0])
        v1 = v0 + self.my_model.dt * (1-self.gamma) * a0

        u1 = u0 + self.my_model.dt * v0 + self.my_model.dt**2 * (0.5-self.beta) * a0
        t1 = t0 + self.my_model.dt
        i1 = i0 + 1

        # compute the internal forces
        r1 = self.my_model.compute_r(u1,v1,self.my_model.g_q[:,i1],t1,i1)
        
        #  f_int =  self.my_model.cumpute_f_int(u1,v1,a1,self.model_g_q[_,i1],t1,i1)

        # residual of the equilibrium equation
        res = self.M @ a1 + self.Cp @ v1 + r1 - self.my_model.Bf @ self.my_model.g_f[:,i1]

        # iteration counter
        cc = 0

        while True:

            # tangent stiffness matrix
            K = self.my_model.compute_K(u1,v1,self.my_model.g_q[:,i1],t1,i1)

            # tangent damping matrix
            C = self.my_model.compute_C(u1,v1,self.my_model.g_q[:,i1],t1,i1)
            
            # Tangent mass matrix
            # M = self.my_model.compute_M(u1,v1,self.my_model.g_q[:,i1],t1,i1)
            
            # jacobian of the residual           
            jac = self.M + (self.Cp + C) * self.gamma * self.my_model.dt + K * self.beta * self.my_model.dt**2

            # acceleration update
            da = -spsolve(jac,res)
 
            # update all kinematic quantities
            a1 = a1 + da
            v1 = v1 + da * self.gamma * self.my_model.dt
            u1 = u1 + da * self.beta  * self.my_model.dt**2

            # compute the internal forces
            r1 = self.my_model.compute_r(u1,v1,self.my_model.g_q[:,i1],t1,i1)

            # residual of the equilibrium equation
            res = self.M @ a1 + self.Cp @ v1 + r1 - self.my_model.Bf @ self.my_model.g_f[:,i1]

            # convergence tolerance
            tol = (np.linalg.norm(res) * self.my_model.ndof) / (self.my_model.ndof * (1 + np.linalg.norm(np.dot(self.my_model.Bf.todense(), self.my_model.g_f[:, i1]))))

            # exit condition
            if cc >= self.iter_max or tol < self.tol_min:
                if output == True:
                    # print(f"step {i1} converged in {cc+1} iterations with residual {tol}")
                    print(simulation.msg.format(i1, cc + 1, tol))
                break
            else:
                cc += 1

        # return the results of the analysis step
        return u1,v1,a1,r1

#%% static analysis
    def static_analysis(self, step1=0, step2=None,output=True):
            
            if step2 is None: 
                step2 = self.my_model.step  # Default last analysis step
            
            # Sequence of analysis steps
            for i in range(step1, step2-1):

                # Single analysis step
                self.u[:, i+1], self.l[:, i+1], self.r[:, i+1] = self.static_analysis_step(self.u[:, i], self.l[:, i], self.t[i], i, output)
        
                # record model results
                self.record_step(i+1)
                
            return self.u, self.l, self.r

    def static_analysis_step(self, u0, l0, t0, i0, output):
        
        u1 = u0.copy()
        l1 = l0.copy()
        i1 = i0 + 1
        t1 = t0 + self.my_model.dt
            
        # Compute the restoring force
        r1 = self.my_model.compute_r(u1, np.zeros_like(u1), self.my_model.g_q[:, i1], t1, i1)

        # Compute the residual of the equilibrium equation  
        res = r1 - self.my_model.Bf @ self.my_model.g_f[:, i1] - self.my_model.Bu @ l1
       
        # Iteration counter
        cc = 0
           
        while True:

            # Jacobian of the residual
            K = self.my_model.compute_K(u1, np.zeros_like(u1), self.my_model.g_q[:, i1], t1, i1)
           
            # Newton-Raphson updates
            duI = -spsolve(K, res)

            # Newton-Raphson updates for Lagrange multipliers
            if self.my_model.Bu.shape[1] != 0 and self.my_model.Bu.nnz != 0:
                g_u_shape = np.shape(self.my_model.g_u)

                if self.my_model.g_u.ndim == 1 or (self.my_model.g_u.ndim == 2 and (self.my_model.g_u.shape[0] == 1 or self.my_model.g_u.shape[1] == 1)):
                    dl = -spsolve(csc_array(np.atleast_2d(self.my_model.Bu.T @ spsolve(K, self.my_model.Bu))), self.my_model.Bu.T @ (u1 + duI) - self.my_model.g_u[:, i1])
                else:
                    dl = -spsolve(self.my_model.Bu.T @ spsolve(K.tocsc(), self.my_model.Bu), self.my_model.Bu.T @ (u1 + duI) - self.my_model.g_u[:, i1])
                
                duII = spsolve(K, self.my_model.Bu @ dl)
            else:
                dl = np.zeros_like(l0)
                duII = np.zeros_like(u0)
            
            # Update all state variables
            u1 = u1 + duI + duII
            l1 = l1 + dl
            
            # Compute the restoring force
            r1 = self.my_model.compute_r(u1, np.zeros_like(u1), self.my_model.g_q[:, i1], t1, i1)

            # Evaluation of the residual
            res = r1 - self.my_model.Bf @ self.my_model.g_f[:, i1] - self.my_model.Bu @ l1

            # Convergence tolerance
            tol = (np.linalg.norm(res) * self.my_model.ndof) / (self.my_model.ndof * (1 + np.linalg.norm(self.my_model.Bf @ self.my_model.g_f[:, i1] + self.my_model.Bu @ l1)))

            # Exit condition
            if cc >= self.iter_max or tol < self.tol_min:
                if output == True:
                    # print(f"step {i1} converged in {cc+1} iterations with residual {tol}")
                    print(simulation.msg.format(i1, cc + 1, tol))
                break
            else:
                # Iteration counter increment
                cc += 1
        
        return u1, l1, r1

#%% Arc-length analysis
    def arclen_analysis(self, step1=0, step2=None, output=True):

        if step2 is None:
            step2 = self.my_model.step

        # Error check
        if self.my_model.dofs_u.shape[0] != 1:
            raise ValueError('arclen analysis supports only 1 displacement controlled dof')

        # Initialize increment direction vectors
        self.Du_last = self.my_model.Bu
        self.Dl_last = 0

        # Sequence of analysis steps
        for i in range(step1, step2-1):

            # Single analysis step
            self.u[:, i+1], self.l[:, i+1], self.r[:, i+1] = self.arclen_analysis_step(self.u[:, i], self.l[:, i], self.t[i], i, output)

            # record model results
            self.record_step(i+1)

        return self.u, self.l, self.r

    def arclen_analysis_step(self, u0, l0, t0, i0, output):
       
        u1 = u0.copy()
        l1 = l0.copy()
        i1 = i0 + 1
        t1 = t0 + self.my_model.dt

        # Initialize increments
        Du = u1 - u0  # displacement increment (total)
        Dl = l1 - l0  # Lagrange multiplier increment (total)

        ds = self.my_model.g_u[:,i1] - self.my_model.g_u[:,i0]  # arc length

        # Compute the restoring force
        r1 = self.my_model.compute_r(u1, np.zeros_like(u1), self.my_model.g_q[:, i1], t1, i1)

        # Compute the residual of the equilibrium equation
        res = r1 - self.my_model.Bf @ self.my_model.g_f[:, i1] - self.my_model.Bu @ l1
        
        # Iteration counter
        cc = 0

        while True:
            # Jacobian of the residual
            K = self.my_model.compute_K(u1, np.zeros_like(u1), self.my_model.g_q[:, i1], t1, i1)

            # Displacement increments predictors.
            duI = spsolve(K, self.my_model.Bu)
            duII = -spsolve(K, res)

            # Lagrange multiplier predictor.
            dl1, dl2 = self.arclen_analysis_dll(duI, duII, Du, Dl, ds)

            if (Du + duII + dl1 * duI) @ self.Du_last > 0:
                dl = dl1.reshape(1,-1)[0]
            else:
                dl = dl2.reshape(1,-1)[0]

            du = duII + dl * duI

            # Update of displacement and Lagrange multipliers (total increments)
            Du += du
            Dl += dl

            # Update of displacement and Lagrange multipliers
            u1 += du
            l1 += dl

            # Compute the restoring force
            r1 = self.my_model.compute_r(u1, np.zeros_like(u1), self.my_model.g_q[:, i1], t1, i1)

            # Compute the residual of the equilibrium equation
            res = r1 - self.my_model.Bf @ self.my_model.g_f[:, i1] - self.my_model.Bu @ l1

            if self.my_model.Bf.shape[1] != 0:
                tol = np.linalg.norm(res) / (1 + sp.sparse.linalg.norm(self.my_model.Bf * self.my_model.g_f[:, i1] + self.my_model.Bu * l1))
            else:
                tol = np.linalg.norm(res) / (1 + sp.sparse.linalg.norm(self.my_model.Bu * l1))

            if cc >= self.iter_max or tol < self.tol_min:
                if output == True:
                    # print(f'step {i1}, iter {cc+1}, res {np.linalg.norm(res)}')
                    print(simulation.msg.format(i1, cc + 1, tol))
                break
            else:
                cc += 1

        # Store the last increment at convergence.
        self.Du_last = Du
        self.Dl_last = Dl

        return u1, l1, r1

    def arclen_analysis_dll(self, duI, duII, Du, Dl, ds):
        h2 = sp.sparse.linalg.norm(self.my_model.Bu)**2
        
        c1 = duI.T @ duI + self.delta**2 * h2
        c2 = 2 * (Du + duII).T @ duI + 2 * self.delta**2 * Dl * h2
        c3 = (Du + duII).T @ (Du + duII) + Dl**2 * self.delta**2 * h2 - ds**2

        discriminant = c2**2 - 4 * c1 * c3
        div_factor = 1 / (2 * c1)

        if discriminant > 0:
            dl1 = (-c2 + np.sqrt(discriminant)) * div_factor
            dl2 = (-c2 - np.sqrt(discriminant)) * div_factor
        else:
            dl1 = -c2 * div_factor
            dl2 = -c2 * div_factor

        return dl1, dl2

#%% store the results of the analysis step from the recorder fields
    def record_step(self,step=0):
        
        # update each field of recorder data
        for recorder_tag in self.recorder_tags:
            self.recorder_data[recorder_tag][step] = eval('self.my_model.' + recorder_tag)
        
        # expand the solution for plots
        self.u_x[self.my_model.ind_x,step] = self.u[:,step]
        self.v_x[self.my_model.ind_x,step] = self.v[:,step]
        self.a_x[self.my_model.ind_x,step] = self.a[:,step]
       
        # store the last step in the model
        self.my_model.u = self.u[:,step]
        self.my_model.v = self.v[:,step]
        self.my_model.a = self.a[:,step]

        # store the last step in the model (expand for plots)
        self.my_model.u_x[self.my_model.ind_x] = self.u[:,step]
        self.my_model.v_x[self.my_model.ind_x] = self.v[:,step]
        self.my_model.a_x[self.my_model.ind_x] = self.a[:,step]

        # dump model results to the elements
        # self.my_model.dump_to_elements()


import numpy as np
from scipy.sparse import coo_array, csr_array
from scipy.linalg import block_diag
from yafem.nodes import nodes
from yafem.elem.core_elem import core_elem
from yafem.elem.plate_func import *
from yafem.elem.solid_func import *

class shell4(core_elem):

    #%% class constructor
    def __init__(self, my_nodes, pars):

        # superclass constructor
        super().__init__(my_nodes,pars)

        self.linear_M = True
        self.linear_K = True
        self.linear_C = True
        
        # extract parameters and assign default values
        self.extract_pars(pars)

        # element dofs
        self.dofs = self.element_dofs(6)
        
        # Calculate the mean of nodal coordinates
        r0 = np.mean(self.nodal_coords, axis=0)
        
        xp = (self.nodal_coords[1, :] + self.nodal_coords[2, :]) / 2 - r0
        xp = xp / np.linalg.norm(xp)

        yp = (self.nodal_coords[2, :] + self.nodal_coords[3, :]) / 2 - r0
        yp = yp - np.dot(yp, xp)
        yp = yp / np.linalg.norm(yp)

        zp = np.cross(xp, yp)

        self.xe = self.nodal_coords
        self.xe_plot = self.xe

        self.xep = np.dot(self.xe - r0, np.column_stack((xp, yp)))
        self.xe = self.xep

        # Local reference system
        self.T = np.array([xp, yp, zp])

        # Transformation matrix for the displacement vector of a single node
        # Assuming BlockDiagonal is a placeholder for an actual block diagonal matrix construction
        # For simplicity, using np.kron (Kronecker product) to simulate block diagonal behavior
        # self.G = np.kron(np.eye(8), self.T)  # Adjust based on actual BlockDiagonal implementation

        self.G = np.zeros((24, 24))
        self.G[0:3, 0:3] = self.T
        self.G[3:6, 3:6] = self.T
        self.G[6:9, 6:9] = self.T
        self.G[9:12, 9:12] = self.T
        self.G[12:15, 12:15] = self.T
        self.G[15:18, 15:18] = self.T
        self.G[18:21, 18:21] = self.T
        self.G[21:24, 21:24] = self.T

        inds = np.array([1, 2])
        indp = np.array([3, -5, 4])
        indd = np.array([6])

        self.Zs = self.fun_mapping(inds)
        self.Zp = self.fun_mapping(indp)
        self.Zd = self.fun_mapping(indd)
                
        d = np.zeros((24, 1))

        self.ds = self.Zs.dot(self.G.dot(d))
        self.dp = self.Zp.dot(self.G.dot(d))
        self.dd = self.Zd.dot(self.G.dot(d))

        self.compute_solid_pars()
        self.compute_plate_pars()
        self.compute_drilling_pars()

        self.Kl  = np.zeros((24, 24))
        self.Ml  = np.zeros((24, 24))
        self.fcl = np.zeros((24, 1))
        self.ftl = np.zeros((24, 1))

        # Local stiffness matrix of the shell element
        self.Kl += self.Zs.T @ self.Kgs @ self.Zs
        self.Kl += self.Zp.T @ self.Kgp @ self.Zp
        self.Kl += self.Zd.T @ self.Kgd @ self.Zd

        # Local mass matrix of the shell element
        self.Ml += self.Zs.T @ self.Mgs @ self.Zs
        self.Ml += self.Zp.T @ self.Mgp @ self.Zp
        self.Ml += self.Zd.T @ self.Mgd @ self.Zd

        # Local consistent nodal load vector
        self.fcl += self.Zs.T @ self.fcgs
        self.fcl += self.Zp.T @ self.fcgp
        self.fcl += self.Zd.T @ self.fcgd

        # Thermo mechanical load vector
        self.ftl += self.Zs.T @ self.ftgs
        self.ftl += self.Zp.T @ self.ftgp
        self.ftl += self.Zd.T @ self.ftgd

        # Element matrices and vectors in global reference system (model)
        self.Kg  = self.G.T @ self.Kl @ self.G
        self.Mg  = self.G.T @ self.Ml @ self.G
        self.fcg = self.G.T @ self.fcl
        self.ftg = self.G.T @ self.ftl
        self.K = self.Kg
        self.M = self.Mg
        self.C = np.zeros(self.K.shape)

    def element_dofs(self, dofs_per_node): 
        self.dofs = np.empty([dofs_per_node*4,2],dtype=int)

        self.dofs[0:dofs_per_node,0]                  = self.nodal_labels[0] # Label of first node
        self.dofs[dofs_per_node*1:dofs_per_node*2,0]  = self.nodal_labels[1] # Label of second node
        self.dofs[dofs_per_node*2:dofs_per_node*3,0]  = self.nodal_labels[2] # Label of third node
        self.dofs[dofs_per_node*3:dofs_per_node*4,0]  = self.nodal_labels[3] # Label of fourth node
        self.dofs[:,1] = np.tile(np.arange(0,dofs_per_node), 4) + 1 # Dofs of all nodes
    
        return self.dofs

    def gl_quadrature(self, n):
        """
        Gauss-Legendre quadrature rule in Python.
        
        :param n: Number of points
        :return: points (s) and weights (w)
        """
        s = {
            1: [0],
            2: [-0.577350, 0.577350],
            3: [-0.774597, 0        , 0.774597],
            4: [-0.861136, -0.339981, 0.339981, 0.861136],
            5: [-0.906179, -0.538469, 0.000000, 0.538469, 0.906179]
            }

        w = {
            1: [2],
            2: [1       , 1],
            3: [5/9     , 8/9     , 5/9],
            4: [0.347855, 0.652145, 0.652145, 0.347855],
            5: [0.236926, 0.478628, 0.568888, 0.478628, 0.236926]
            }

        return s[n], w[n]
    
    def polyarea(self, xe):
        # xe = np.asarray(xe)
        x = xe[:, 0]
        y = xe[:, 1]
        x_next = np.concatenate((x[1:], [x[0]]))  # Shift x-coordinates
        y_next = np.concatenate((y[1:], [y[0]]))  # Shift y-coordinates
        A2 = np.sum(x * y_next - x_next * y)
        return abs(A2 / 2)


    #%% Mapping
    
    def fun_mapping(self, ind):
        Z = np.zeros((len(ind), 6))

        Z[range(len(ind)), abs(ind) - 1] = np.sign(ind)
        Z = block_diag(Z,Z,Z,Z)
                      
        return coo_array(Z,dtype=np.int8).tocsr()

    #%% Solid behavior
    def compute_solid_pars(self, ds=None):
        """
        Compute the solid parameters for the element.
        """
        if ds is None:
            ds = np.zeros((8, 1))

        # Assuming e_fun_solid is an object or module that contains the functions Dps, Dpe, and Dax
        type_to_D = {
            'ps': solid_Dps,
            'pe': solid_Dpe,
            'ax': solid_Dax
        }

        # Assuming 'e' is an object that has attributes 'type', 'E', and 'nu'
        # Attempt to get the function based on e.type, default to raising an error if not found
        if self.type in type_to_D:
            D = type_to_D[self.type]
            self.D = D(self.E, self.nu)  # Call the function with e.E and e.nu as arguments
        else:
            raise ValueError("Type is not specified correctly!")
 
        sf, wf = self.gl_quadrature(2)
        sr, wr = self.gl_quadrature(1)

        self.Kgs  = np.zeros((8, 8))
        self.Mgs  = np.zeros((8, 8))
        self.fcgs = np.zeros((8, 1))
        # Initialize 2D arrays to store strain and stress
        self.eps_n = np.empty((2, 2), dtype=object)
        self.sig_n = np.empty((2, 2), dtype=object)
        self.eps_s = np.empty((1, 1), dtype=object)
        self.sig_s = np.empty((1, 1), dtype=object)
        
        if self.type == "ax":
            # Gauss-Legendre integration (full)
            sf, wf = self.gl_quadrature(2)
            for i in range(len(sf)):
                for j in range(len(sf)):
                    # Jacobian of the iso-parametric mapping in the point ij
                    Jac = solid_Jac(sf[i], sf[j], self.xe).reshape((2, 2))

                    # Strain interpolation matrix
                    N = solid_N(sf[i], sf[j])

                    # Shape functions
                    phi = solid_phi(sf[i], sf[j])

                    # Density
                    rho_ij = np.dot(phi, self.xe[:, 1])

                    inv_Jac = np.linalg.inv(Jac)    
                    # inv_Jac_blc = block_diag(inv_Jac, inv_Jac)
                    inv_Jac_blc = np.zeros((4, 4))
                    inv_Jac_blc[0:2, 0:2] = inv_Jac
                    inv_Jac_blc[2:4, 2:4] = inv_Jac

                    # Strain interpolation matrix
                    B = np.vstack([
                        np.dot(np.array([[1, 0, 0, 0], [0, 0, 0, 1]]), inv_Jac_blc.T) @ solid_B(sf[i], sf[j]),
                        np.dot(np.array([1, 0]), N) / rho_ij
                    ])

                    # Done once per gauss point
                    dJ = np.linalg.det(Jac) * wf[i] * wf[j] * rho_ij
                    
                    if dJ < 0:
                        raise('negative determinant of the jacobian')

                    # Gauss-Legendre sum for stiffness
                    self.Kgs += B.T @ self.D[0:3, 0:3] @ B * dJ * 2 * np.pi

                    # Gauss-Legendre sum for mass
                    self.Mgs += N.T @ N * self.rho * dJ * 2 * np.pi

                    # strain (e_rr,e_yy,e_tt)
                    self.eps_n[i, j] = B @ ds

                    # stress (e_rr,e_yy,e_tt)
                    self.sig_n[i, j] = self.D[0:3, 0:3] @ self.eps_n[i, j]

                    # Gauss-Legendre sum for distributed load vector
                    self.fcgs += (N.T @ np.array([self.bx, self.by]).reshape(-1, 1)) * dJ * 2 * np.pi

            # Gauss-Legendre integration (reduced)
            sr, wr = self.gl_quadrature(1)
            for i in range(len(sr)):
                for j in range(len(sr)):
                    # Jacobian of the iso-parametric mapping in the point ij
                    Jac = solid_Jac(sr[i], sr[j], self.xe).reshape((2, 2))

                    # Strain interpolation matrix
                    N = solid_N(sr[i], sr[j])

                    # Shape functions
                    phi = solid_phi(sr[i], sr[j])

                    # Density
                    rho_ij = np.dot(phi, self.xe[:, 1])

                    inv_Jac = np.linalg.inv(Jac)    
                    # inv_Jac_blc = block_diag(inv_Jac, inv_Jac)
                    inv_Jac_blc = np.zeros((4, 4))
                    inv_Jac_blc[0:2, 0:2] = inv_Jac
                    inv_Jac_blc[2:4, 2:4] = inv_Jac


                    # Strain interpolation matrix
                    B = np.dot(np.array([0, 1, 1, 0]), inv_Jac_blc.T) @ solid_B(sr[i], sr[j])

                    # Done once per gauss point
                    dJ = np.linalg.det(Jac) * wr[i] * wr[j] * rho_ij
                    
                    if dJ < 0:
                        raise('negative determinant of the jacobian')
                        
                    # Gauss-Legendre sum for stiffness
                    self.Kgs += B.T @ B * self.D[3, 3] * dJ * 2 * np.pi

                    # strain (e_rr,e_yy,e_tt)
                    self.eps_s[i, j] = B @ ds

                    # stress (e_rr,e_yy,e_tt)
                    self.sig_s[i, j] = self.D[3, 3] * self.eps_s[i, j]
        else:
            # Gauss-Legendre full
            sf, wf = self.gl_quadrature(2)  # Assuming sf and wf are defined similarly to sr and wr
            for i in range(len(sf)):
                for j in range(len(sf)):
                    # Jacobian of the iso-parametric mapping in the point ij
                    Jac = solid_Jac(sf[i], sf[j], self.xe).reshape((2, 2))

                    # Strain interpolation matrix
                    N = solid_N(sf[i], sf[j])

                    # Shape functions
                    phi = solid_phi(sf[i], sf[j])

                    # Density
                    rho_ij = np.dot(phi, self.xe[:, 1])
                    Jac_inv = np.linalg.inv(Jac)
                    # Jac_inv_blc = block_diag(Jac_inv, Jac_inv)
                    Jac_inv_blc = np.zeros((4, 4))
                    Jac_inv_blc[0:2, 0:2] = Jac_inv
                    Jac_inv_blc[2:4, 2:4] = Jac_inv


                    # Strain interpolation matrix
                    B = np.dot(np.array([[1, 0, 0, 0], [0, 0, 0, 1]]), Jac_inv_blc.T) @ solid_B(sf[i], sf[j])

                    # Done once per gauss point
                    dJ = np.linalg.det(Jac) * wf[i] * wf[j]
                    
                    if dJ < 0:
                        raise('negative determinant of the jacobian')                    

                    # Gauss-Legendre sum for stiffness
                    self.Kgs += B.T @ self.D[0:2, 0:2] @ B * dJ * self.h

                    # Gauss-Legendre sum for mass
                    self.Mgs += N.T @  N * self.rho * dJ * self.h

                    # strain (e_rr,e_yy,e_tt)
                    self.eps_n[i, j] = B @ ds

                    # stress (e_rr,e_yy,e_tt)
                    self.sig_n[i, j] = self.D[0:2, 0:2] @ self.eps_n[i, j]

                    # Gauss-Legendre sum for distributed load vector

                    # Perform the matrix multiplication, then reshape the result to (8, 1) before multiplying by dJ
                    self.fcgs += (N.T @ np.array([self.bx, self.by]).reshape(-1, 1)) * dJ


            # Gauss-Legendre integration (reduced)
            sr, wr = self.gl_quadrature(1)
            for i in range(len(sr)):
                for j in range(len(sr)):
                    # Jacobian of the iso-parametric mapping in the point ij
                    Jac = solid_Jac(sr[i], sr[j], self.xe).reshape((2, 2))

                    # Strain interpolation matrix
                    N = solid_N(sr[i], sr[j])

                    # Shape functions
                    phi = solid_phi(sr[i], sr[j])

                    # Density
                    rho_ij = np.dot(phi, self.xe[:, 1])
                    Jac_inv = np.linalg.inv(Jac)
                    
                    # Assuming the use of block_diag is necessary as in the full integration example
                    # If not applicable here, adjust accordingly
                    # Jac_inv_blc = block_diag(Jac_inv, Jac_inv)
                    Jac_inv_blc = np.zeros((4, 4))
                    Jac_inv_blc[0:2, 0:2] = Jac_inv
                    Jac_inv_blc[2:4, 2:4] = Jac_inv

                    # Strain interpolation matrix
                    B = np.dot(np.array([[0, 1, 1, 0]]), Jac_inv_blc.T) @ solid_B(sr[i], sr[j])

                    # Done once per gauss point
                    dJ = np.linalg.det(Jac) * wr[i] * wr[j]
                    
                    if dJ < 0:
                        raise('negative determinant of the jacobian')                    

                    # Gauss-Legendre sum for stiffness
                    self.Kgs += B.T @  B * self.D[2, 2] * dJ * self.h

                    # Before the operation, print out the shapes of N, self.rho, dJ, and self.h


                    # Gauss-Legendre sum for mass
                    # self.Mgs += N.T @ N * self.rho * dJ * self.h

                    # strain (e_rr,e_yy,e_tt)
                    self.eps_s[i, j] = B @ ds

                    # stress (e_rr,e_yy,e_tt)
                    self.sig_s[i, j] = self.D[2, 2] * self.eps_s[i, j]
        self.ftgs = np.zeros(self.fcgs.shape)
        self.ds = ds
       
    #%% Plate behavior
    def compute_plate_pars(self, dp=None):
        
        if dp is None:
            dp = np.zeros((12, 1))
        
        self.Dbp = plate_Db(self.E, self.nu) * self.h**3 / 12
        self.Dsp = plate_Ds(self.E, self.nu) * self.h

        # Gauss-Legendre integration (full and reduced)
        sf, wf = self.gl_quadrature(2)
        sr, wr = self.gl_quadrature(1)

        self.Kb = np.zeros((12, 12))
        
        self.defb = np.empty((2, 2), dtype=object)
        self.forb = np.empty((2, 2), dtype=object)

        Bb = np.empty(plate_Bb(1, 1).shape, dtype=float)
        
        Jac = np.empty((2, 2), dtype=float)

        # Computing bending stiffness for bending and twisting
        for i in range(len(sf)):
            for j in range(len(sf)):
                Jac = plate_Jac(sf[i], sf[j], self.xe).reshape((2, 2))
                Bb  = plate_Bb(sf[i], sf[j])
                
                # Blockdiagonal of Jac
                Jacblk = block_diag(Jac, Jac)

                B = np.dot(np.array([[1, 0, 0, 0], [0, 0, 0, 1], [0, 1, 1, 0]]), np.linalg.solve(Jacblk, Bb))

                dJ = np.linalg.det(Jac) * wf[i] * wf[j]
                
                if dJ < 0:
                    raise('negative determinant of the jacobian')                

                self.Kb += B.T @ self.Dbp @ B * dJ

                self.defb[i, j] = B @ dp
                self.forb[i, j] = self.Dbp @ self.defb[i, j]

        # Computing the mass stiffness matrix
        self.Mb = np.zeros((12, 12))

        N = np.empty(plate_N(1, 1).shape, dtype=float)
        Jac = np.empty((2, 2), dtype=float)

        for i in range(len(sf)):
            for j in range(len(sf)):
                Jac = plate_Jac(sf[i], sf[j], self.xe).reshape((2, 2))
                N = plate_N(sf[i], sf[j])
                dJ = np.linalg.det(Jac) * wf[i] * wf[j]

                # Create the diagonal matrix
                diagonal_matrix = np.diag([self.h, self.h**3/12, self.h**3/12])

                # Update self.Mb using the corrected formula
                self.Mb += N.T @ (self.rho * diagonal_matrix) @ N * dJ

        # Shear stiffness matrix full integration
        self.Ks2 = np.zeros((12, 12))

        Jac = np.empty((2, 2), dtype=float)
        Bs1 = np.empty(plate_Bs1(1, 1).shape, dtype=float)
        Bs2 = np.empty(plate_Bs2(1, 1).shape, dtype=float)

        for i in range(len(sf)):
            for j in range(len(sf)):
                Jac = plate_Jac(sf[i], sf[j], self.xe).reshape((2, 2))
                Bs1 = plate_Bs1(sf[i], sf[j])
                Bs2 = plate_Bs2(sf[i], sf[j])

                Bs = Bs1 + np.linalg.solve(Jac, Bs2)

                dJ = np.linalg.det(Jac) * wf[i] * wf[j]
                
                if dJ < 0:
                    raise('negative determinant of the jacobian')
                        
                self.Ks2 += Bs.T @ self.Dsp @ Bs * dJ

        # Shear stiffness matrix reduced integration
        self.Ks1 = np.zeros((12, 12))
        self.defs = np.empty((1, 1), dtype=object)
        self.fors = np.empty((1, 1), dtype=object)

        Jac = np.empty((2, 2), dtype=float)
        Bs1 = np.empty(plate_Bs1(1, 1).shape, dtype=float)
        Bs2 = np.empty(plate_Bs2(1, 1).shape, dtype=float)

        for i in range(len(sr)):
            for j in range(len(sr)):
                Jac = plate_Jac(sr[i], sr[j], self.xe).reshape((2, 2))
                Bs1 = plate_Bs1(sr[i], sr[j])
                Bs2 = plate_Bs2(sr[i], sr[j])

                Bs = Bs1 + np.linalg.solve(Jac, Bs2)

                dJ = np.linalg.det(Jac) * wr[i] * wr[j]
                
                if dJ < 0:
                    raise('negative determinant of the jacobian')
                
                self.Ks1 += Bs.T @ self.Dsp @ Bs * dJ

                self.defs[i, j] = Bs @ dp
                self.fors[i, j] = self.Dsp @ self.defs[i, j]

        # Stiffness matrix total
        # Hourglass control
        poly_area = self.polyarea(self.xe)
        self.delta = self.epsilon * self.h**2 / poly_area
        self.Ks = self.Ks1 + self.delta * (self.Ks2 - self.Ks1)
        self.Kp = self.Kb + self.Ks

        # Calculate the force vector and element matrices
        self.fcp = np.zeros((12, 1))
        N = np.empty(plate_N(1, 1).shape, dtype=float)
        Jac = np.empty((2, 2), dtype=float)

        for i in range(len(sf)):
            for j in range(len(sf)):
                Jac = plate_Jac(sf[i], sf[j], self.xe).reshape((2, 2))
                N = plate_N(sf[i], sf[j])

                dJ = np.linalg.det(Jac) * wf[i] * wf[j]
                
                if dJ < 0:
                    raise('negative determinant of the jacobian')

                self.fcp += N.T @ np.array([self.bz, 0, 0]).reshape(-1, 1) * dJ

        self.Kgp = self.Kp
        self.Mgp = self.Mb
        self.fcgp = self.fcp
        self.ftgp = np.zeros(self.fcp.shape)

    #%% Drilling behavior

    def compute_drilling_pars(self):

        k3 = np.array([[   1, -1/2, -1/2], 
                       [-1/2,    1, -1/2], 
                       [-1/2, -1/2,    1]])
                
        k4a = np.zeros((4, 4))
        k4b = np.zeros((4, 4))
        
        poly_area_012 = self.polyarea(self.xe[[0,1,2], :])  
        poly_area_123 = self.polyarea(self.xe[[1,2,3], :])
        poly_area_013 = self.polyarea(self.xe[[0,1,3], :])
        poly_area_023 = self.polyarea(self.xe[[0,2,3], :])

        k4a[np.ix_([0, 1, 2], [0, 1, 2])] = k3 * self.h * self.E * self.alpha * poly_area_012
        k4a[np.ix_([1, 2, 3], [1, 2, 3])] = k4a[np.ix_([1, 2, 3], [1, 2, 3])] + k3 * self.h * self.E * self.alpha * poly_area_123
        
        # k4b[np.ix_([0, 2, 3], [0, 2, 3])] = k3 * self.h * self.E * self.alpha * poly_area_3
        # k4b[np.ix_([0, 1, 3], [0, 1, 3])] = k4b[np.ix_([0, 1, 3], [0, 1, 3])] + k3 * self.h * self.E * self.alpha * poly_area_4
          
        k4b[np.ix_([0, 2, 3], [0, 2, 3])] = k3 * self.h * self.E * self.alpha * poly_area_023
        k4b[np.ix_([0, 1, 3], [0, 1, 3])] = k4b[np.ix_([0, 1, 3], [0, 1, 3])] + k3 * self.h * self.E * self.alpha * poly_area_013
      
        self.Kd   = (k4a + k4b) / 2
        self.Kgd  = self.Kd
        self.fcgd = np.zeros((4, 1))
        self.ftgd = np.zeros((4, 1))
        self.Mgd  = self.Kgd/1e10


    #%% Extract parameters

    def extract_pars(self, pars):
        # this is the element class used in packing/unpacking
        self.my_pars['elem'] = 'shell4'

        self.E   = pars.get("E", 210e9)
        self.nu  = pars.get("nu", 0.3)
        self.rho = pars.get("rho", 7850)
        self.h   = pars.get("h", 5e-3)
        self.nodal_labels = pars.get("nodal_labels", [1, 2, 3, 4]) # node labels
        self.nodal_coords = self.my_nodes.find_coords(self.nodal_labels) # extract nodal coordinates
        self.alpha   = pars.get("alpha", 0.3)
        self.epsilon = pars.get("epsilon", 0.01)
        self.bx      = pars.get("bx", 0.0)
        self.by      = pars.get("by", 0.0)
        self.bz      = pars.get("bz", 0.0)
        self.type    = pars.get("type", "ps")  # type of analysis (ps = plane stress, pe = plane strain, ax = axisymmetric)
        self.dofs_q  = pars.get("dofs_q", np.zeros((0, 2), dtype=np.int32))
        
    #%% Plot 3d elements
    def plot(self, ax, x=None, y=None, z=None, color='cyan'):
        if x is None: x = self.nodal_coords[:, 0]
        if y is None: y = self.nodal_coords[:, 1]
        if z is None: z = self.nodal_coords[:, 2]
            
        # Use nodal coordinates directly to form a quadrilateral grid
        surfaces = [[[x[0], y[0], z[0]], 
                     [x[1], y[1], z[1]], 
                     [x[2], y[2], z[2]], 
                     [x[3], y[3], z[3]]]]

        # Collect lines
        lines = [[[x[0], y[0], z[0]], [x[1], y[1], z[1]]],
                 [[x[1], y[1], z[1]], [x[2], y[2], z[2]]],
                 [[x[2], y[2], z[2]], [x[3], y[3], z[3]]],
                 [[x[3], y[3], z[3]], [x[0], y[0], z[0]]]]

        return lines, surfaces

    def dump_to_paraview(self):
        # here it goes the dump_to_paraview implementation for the beam3d element
        pass


import numpy as np
import jax
import jax.numpy as jnp
from jax import jacrev, jit
jax.config.update("jax_enable_x64", True)
from yafem.nodes import nodes
from yafem.elem.core_elem import core_elem
from dataclasses import dataclass

#%% element_MCK class
class beam2d_cr(core_elem):

    # class constructor
    def __init__(self, my_nodes, pars):

        # superclass constructor
        super().__init__(my_nodes,pars)

        self.linear_M = False
        self.linear_K = False
        self.linear_C = False

        # extract parameters and assign default values
        self.extract_pars(pars)

        # element dofs
        self.dofs = np.array([[self.nodal_labels[0],1],   
                              [self.nodal_labels[0],2],   
                              [self.nodal_labels[0],3],   
                              [self.nodal_labels[1],1],
                              [self.nodal_labels[1],2],
                              [self.nodal_labels[1],3]],dtype=np.int32)
        
        # jax variables
        # u = jnp.array()
        # v = jnp.array()
        # a = jnp.array()
        # q = jnp.array()
        # t = jnp.array()
        # i = jnp.array()

        # nodal coordinates
        x1 = self.nodal_coords[0,0]
        x2 = self.nodal_coords[1,0]
        z1 = self.nodal_coords[0,1]
        z2 = self.nodal_coords[1,1]

        #%% fixed numbers

        # inition length
        l0 = ((x2-x1)**2 + (z2-z1)**2)**0.5 # E6 #### fixed

        # cosine parameters
        c0 = (x2-x1)/l0 #E8 #### fixed
        s0 = (z2-z1)/l0 #E9 #### fixed

        beta0 = jnp.atan2(s0,c0) #### fixed

        # axial stiffness
        ka = self.E*self.A/l0 #### fixed

        # bending stiffness
        kb = 6*self.E*self.I/l0**2 #### fixed

        # local stiffness matrix
        Kl = jnp.array([[ka ,  0  , 0 ],
                        [0  ,  kb ,-kb],
                        [0  , -kb , kb]]) #### fixed
        
        Ml2 = self.rho*self.I/(30*l0) * jnp.array([[0 , 0    , 0       ,0  ,  0    , 0       ] ,
                                                   [0 , 36   , 3*l0    ,0  , -36   , 3*l0    ] ,
                                                   [0 , 3*l0 , 4*l0**2 ,0  , -3*l0 ,-l0**2   ] ,
                                                   [0 , 0    , 0       ,0  ,  0    , 0       ] ,
                                                   [0 ,-36   ,-3*l0    ,0  ,  36   ,-3*l0    ] ,
                                                   [0 , 3*l0 ,-l0**2   ,0  , -3*l0 , 4*l0**2 ]]) # E45B #### fixed

        I1 = jnp.array([[ 0 ,1 ,0, 0 ,0 ,0],
                        [-1 ,0 ,0, 0 ,0 ,0],
                        [ 0 ,0 ,0, 0 ,0 ,0],
                        [ 0 ,0 ,0, 0 ,1 ,0],
                        [ 0 ,0 ,0,-1 ,0 ,0],
                        [ 0 ,0 ,0, 0 ,0 ,0]]) #E53 #### fixed
        
        Mt1bl = self.rho*l0*self.A/60*jnp.array([[ 0 , 3 , 0 , 0 ,-3 , 0],
                                                 [ 3 , 0 , 0 , 2 , 0 , 0],
                                                 [ 0 , 0 , 0 , 0 , 0 , 0],
                                                 [ 0 , 2 , 0 , 0 ,-2 , 0],
                                                 [-3 , 0 , 0 ,-2 , 0 , 0],
                                                 [ 0 , 0 , 0 , 0 , 0 , 0]]) #E55 #### fixed

        Mt2bl = self.rho*l0*self.A/60*jnp.array([[ 0 ,-2 , 0 , 0 , 2 , 0],
                                                 [-2 , 0 , 0 ,-3 , 0 , 0],
                                                 [ 0 , 0 , 0 , 0 , 0 , 0],
                                                 [ 0 ,-3 , 0 , 0 , 3 , 0],
                                                 [ 2 , 0 , 0 , 3 , 0 , 0],
                                                 [ 0 , 0 , 0 , 0 , 0 , 0]]) #56 #### fixed

        self.cr_params = (
            x1,
            x2,
            z1,
            z2,
            l0,
            beta0,
            self.E,
            self.A,
            self.I,
            self.rho,
            Kl,
            Ml2,
            I1,
            Mt1bl,
            Mt2bl
            )

        u = np.zeros_like(self.nodal_coords.flatten())
        v = np.zeros_like(u)
        a = v.copy()

        self.f = self.compute_f(u,v,a)
        self.M = self.compute_M(u,v,a)
        self.C = self.compute_C(u,v,a)
        self.K = self.compute_K(u,v,a)


        #%% updated numbers

    def compute_f_jax(u, v, a, cr_params):
        (x1, x2, z1, z2, l0, beta0, E, A, I, rho, Kl, Ml2, I1, Mt1bl, Mt2bl) = cr_params

        # Unpack inputs
        u1, w1, t1, u2, w2, t2 = u
        u1d, w1d, t1d, u2d, w2d, t2d = v
        u1dd, w1dd, t1dd, u2dd, w2dd, t2dd = a
        
         # current length
        ln = (((x2+u2)-(x1+u1))**2 + ((z2+w2)-(z1+w1))**2)**0.5 #E7
        
        # cosine parameters
        c = (x2+u2-x1-u1)/ln #E8
        s = (z2+w2-z1-w1)/ln #E9

        beta  = jnp.atan2(s,c)

        z = jnp.array([ s,-c, 0,-s, c, 0])

        # local displacements (scalar)
        ub  = ln-l0
        t1b = t1 - beta - beta0
        t2b = t2 - beta - beta0

        ql = jnp.array([ub,t1b,t2b])

        # deformation interpolation matrix
        b1 = jnp.array([-c   ,-s   , 0,    c,    s, 0]) #E12
        b2 = jnp.array([-s/ln, c/ln, 1, s/ln,-c/ln, 0]) #E12
        b3 = jnp.array([-s/ln, c/ln, 0, s/ln,-c/ln, 1]) #E12

        B = jnp.array([b1,b2,b3]) #E12
        
        # local restoring force vector
        fl = Kl @ ql

        # global restoring force vector
        rg = B.T @ fl # E14

        T = jnp.array([[ c,s,0, 0,0,0],
                       [-s,c,0, 0,0,0],
                       [ 0,0,1, 0,0,0],
                       [ 0,0,0, c,s,0],
                       [ 0,0,0,-s,c,0],
                       [ 0,0,0, 0,0,1]])
        
        m1 = 21*t1b-14*t2b
        m2 = 14*t1b-21*t2b

        Ml1 = rho*A*l0/420 * jnp.array([[ 140 ,  m1    , 0       , 70  ,-m1    , 0       ] ,
                                        [ m1  ,  156   , 22*l0   , m2  , 54    ,-13*l0   ] ,
                                        [ 0   ,  22*l0 , 4*l0**2 , 0   , 13*l0 ,-3*l0**2 ] ,
                                        [ 70  ,  m2    , 0       , 140 ,-m2    , 0       ] ,
                                        [-m1  ,  54    , 13*l0   , -m2 , 156   ,-22*l0   ] ,
                                        [ 0   , -13*l0 ,-3*l0**2 , 0   , -22*l0, 4*l0**2 ]]) #E45A 

        Ml = Ml1 + Ml2 # E45
        M = T.T @ Ml @ T # E43
        
        Mbeta = T.T @ (I1.T @ Ml + Ml @ I1) @ T # E54

        Mt1b = T.T @ Mt1bl @ T #E55
        Mt2b = T.T @ Mt2bl @ T #E56

        Md = Mbeta * (z @ v)/ln + Mt1b*(b2 @ v) + Mt2b*(b3 @ v) #E49

        rk = M @ a + Md @ v - 1/2 * (v @ Mbeta @ v) * z/ln - 1/2 * (v @ Mt1b @ v) * b2 - 1/2 * (v @ Mt2b @ v) * b3 #E51

        r = rg + rk

        return r

    # def compute_M(self,u,v,q,t,i): # Check the inputs here
    #     pass

    # def compute_C(self,u,v,q,t,i): # Check the inputs here
    #     pass

    # def compute_K(self,u,v,q,t,i): # Check the inputs here
    #     pass

    def compute_f(self,u,v,a,q=None,t=None,i=None): # Check the inputs here
        f_jit = jit(beam2d_cr.compute_f_jax)
        f = f_jit(u,v,a,self.cr_params)
        return f

    def compute_M(self,u,v,a,q=None,t=None,i=None): # Check the inputs here
        M_jit = jit(jacrev(beam2d_cr.compute_f_jax,argnums=2))
        M = M_jit(u,v,a,self.cr_params)
        return M

    def compute_C(self,u,v,a,q=None,t=None,i=None): # Check the inputs here
        C_jit = jit(jacrev(beam2d_cr.compute_f_jax,argnums=1))
        C = C_jit(u,v,a,self.cr_params)
        return C

    def compute_K(self,u,v,a,q=None,t=None,i=None): # Check the inputs here
        K_jit = jit(jacrev(beam2d_cr.compute_f_jax,argnums=0))
        K = K_jit(u,v,a,self.cr_params)
        return K

    #%% extract parameters and assign default values
    def extract_pars(self,pars):

        self.E   = pars.get('E', 210e3)
        self.I   = pars.get('I', 1000)
        self.A   = pars.get('A', 200)
        self.rho = pars.get('rho', 7850)

        # this is the element class used in packing/unpacking
        self.my_pars['elem'] = 'beam2d_cr'


        #%% plot the element       
    def plot(self, ax, x=None, y=None, z=None, color='k-'):
        if x is None: x = self.nodal_coords[:, 0]
        if y is None: y = self.nodal_coords[:, 1]
        if z is None: z = self.nodal_coords[:, 2]

        # Collect lines
        lines = [[[x[0], y[0], 0],[x[1], y[1], 0]]]

        return lines, None




import numpy as np
import matplotlib.pyplot as plt
import matplotlib.patches as patches
from scipy.linalg import block_diag
from yafem.nodes import nodes
from yafem.elem.core_elem import core_elem
from yafem.elem.beam3d_circ_func import *
from yafem.elem.beam3d_gen_func import *
import functools

def disabled_method(method):
    """Decorator to disable a method in a subclass so that the superclass method is used."""
    @functools.wraps(method)
    def wrapper(*args, **kwargs):
        raise NotImplementedError("This method is disabled and should not be used.")
    
    wrapper._disabled = True  # Custom attribute to mark as disabled
    return wrapper

class userdef(core_elem):
    #%% class constructor
    def __init__(self, my_nodes, pars):

        # superclass constructor
        super().__init__(my_nodes,pars)

        # link the nodes to the element
        self.my_nodes = my_nodes
        
        # extract parameters and assign default values
        self.extract_pars(pars)

        # Stiffness matrix in global coordinate system
        self.K = np.zeros((1,1))
        
        # Mass matrix
        self.M = np.zeros((1,1))

        # damping matrix in global coordinates
        self.C = np.zeros((1,1))

        # local to global coordinate transformation
        self.r = np.zeros((1))
        
    #%% extract parameters
    def extract_pars(self, pars):
        
        # this is an example of how to extract a parameter and assign a default value
        self.example = pars.get("example",0)
 
    #%% compute the mass matrix
    @disabled_method # this is already defined like this in core_elem; so keep it only if you need to edit it
    def compute_M(self):
        return self.M

    #%% compute the damping matrix
    @disabled_method # this is already defined like this in core_elem; so keep it only if you need to edit it
    def compute_C(self,u,v,q,t,i):
        return self.C

    #%% compute the stiffness matrix
    @disabled_method # this is already defined like this in core_elem; so keep it only if you need to edit it
    def compute_K(self,u,v,q,t,i):
        return self.K

    #%% compute the restoring force
    @disabled_method # this is already defined like this in core_elem; so keep it only if you need to edit it
    def compute_r(self,u,v,q,t,i):

        # compute the restoring force
        r = self.K @ u + self.C @ v

        # store the state variables
        self.u = u
        self.v = v
        self.r = r
        self.t = t 
        self.i = i

        # return the restoring force
        return r
     
    #%% add the element plot to the axis ax
    @disabled_method # this is already defined like this in core_elem; so keep it only if you need to edit it
    def plot(self, ax, x=None, y=None, z=None, color='k-'):
        pass
    
    #%% damp the element results to paraview file
    @disabled_method # this is already defined like this in core_elem; so keep it only if you need to edit it
    def dump_to_paraview(self):
        # here it goes the dump_to_paraview implementation for the beam3d element
        pass



#%%
import shutil
from setuptools import setup
from setuptools import find_packages
import os

#%% Remove old build directory
build_folder = 'build'
if os.path.exists(build_folder):
    shutil.rmtree(build_folder)

#%% Input section
package_name = 'yafem'
description  = 'Yet Another Finite Element Method (YAFEM) Package'
version      = '1.0.0'
requirements = False 

#%% Load requirements
if requirements == True:
    with open('requirements.txt') as f:
        required = f.read().splitlines()
else:
    required = ('numpy',
                'scipy',
                'jax',
                'jaxlib',
                'matplotlib',
                )

# # Generating license file of the "required"
# generate_license(required)

#%% Exclude redundant files

# all files in the package_name path
yafem_files = os.listdir(package_name + '\\elem')

# dev files
yafem_dev_files = [s for s in yafem_files if "_dev" in s]

# file to generate lamdified functions
yafem_functions = [s for s in yafem_files if "fun_" in s]

# concatinated files to be ignored when packing
exclude_files = yafem_dev_files + yafem_functions + ["testing", "testing.*"]

#%% Setup

setup(
    # setup_requires=['wheel'],
    name = package_name,
    version = version,
    description = description,
    long_description= open('readme_package_description.md').read(),
                    #   + ' \n# Change log of version ' + version + ': \n\n' + \
                    #   open('readme_change_log\\readme_changes_' + version + '.md').read(),             
    install_requires = required,
    packages = find_packages(exclude = exclude_files),
    # package_data={'': ['LICENSE.md','NOTICE.md']},
    package_dir={"yafem": "yafem"},
    license="MIT",            
    license_files=["LICENSE", "NOTICE"],
    )


# #%% Creating a requirements file

# dist_files = os.listdir('dist')
# yafem_wheel = [s for s in dist_files if ".whl" in s and version in s]

# with open('dist\\requirements.txt', 'w') as f:
#     f.write(yafem_wheel[0] + '\n')

# with open('dist\\requirements.txt', 'a') as f:
#     f.write('sympy' + '\n'
#             'numpy' + '\n'
#             'scipy' + '\n'
#             'pandas' + '\n'
#             'jax' + '\n'
#             'jaxlib' + '\n'
#             'matplotlib' + '\n'
#             'ipykernel' + '\n'
#             'mkdocs-material' + '\n'
#             'pipdeptree' + '\n'
#             'gmsh' + '\n'
#             'raschii' + '\n'
#             )


import numpy as np
import scipy as sp
import concurrent.futures
import jax
import jax.numpy as jnp
import matplotlib.pyplot as plt
import matplotlib.patches as patches
import scipy.sparse.linalg as spla
from scipy.sparse import eye as speye
from scipy.sparse import csr_array, coo_array, csc_array
from scipy.sparse.linalg import eigsh
from yafem.nodes import *
from yafem.elem.core_elem import core_elem
from mpl_toolkits.mplot3d.art3d import Line3DCollection, Poly3DCollection

# Caches
cache_eigenmodes = {} # Cache for eigenvectors
# cache_collocations = {} # Cache for collocations
# cache_M = {} # Cache for mass matrix
# cache_K = {} # Cache for stiffness matrix

class model:
#%% this is the class constructor
    def __init__(self, my_nodes, my_elements, pars):

        # Assumed the model is model is non-linear in M, C and K 
        # so the model is assembled at least once
        self.linear_M = False
        self.linear_C = False
        self.linear_K = False

        # save the link the object my_nodes
        self.my_nodes = my_nodes

        # save the link the list object list my_elements
        self.my_elements = my_elements

        # extract the parameters and assign default values
        self.extract_pars(pars)

        # list of model dofs
        self.dofs = np.zeros((0, 2), dtype=int)

        # assembly of the list of model dofs
        for my_element in self.my_elements:

            # for each dof of the element
            for idx_e, dof_e in enumerate(my_element.dofs):
                # loop over the slave dofs
                for idx_s, dof_s in enumerate(self.dofs_s):
                    # if the element dof is a slave dof
                    if np.array_equal(dof_e, dof_s):
                        # replace the element slave dof with the corresponding master dof
                        my_element.dofs[idx_e] = self.dofs_m[idx_s]

            # add the element dofs to the model dofs
            self.dofs = np.concatenate((self.dofs, my_element.dofs), axis=0)

        # remove repeated dofs
        self.dofs = np.unique(self.dofs, axis=0)

        # Removal of constraint dofs (dofs_c)
        if jnp.shape(self.dofs_c)[0] != 0:
            @jax.jit
            def collocation_indices(dof_c, dofs):

                def indices(dof_c, dofs):
                    match = jnp.all(dofs == dof_c, axis=1)
                    return jnp.argmax(match)

                idx = jax.vmap(indices, in_axes=(0, None))(dof_c, dofs)  

                return idx            

            idx = collocation_indices(self.dofs_c, self.dofs)
            self.dofs = np.delete(self.dofs, idx, axis=0)          

        # compute the collocation matrices of all elements
        self.compute_collocations()

        # if there are master-slave equations
        if self.dofs_m.shape[0] > 0:
            # master-slave matrix
            self.Tu = speye(self.dofs.shape[0]).todense()
            # link slave to master dofs
            self.Tu[self.find_dofs(self.dofs_s), self.find_dofs(self.dofs_m)] = speye(self.dofs_m.shape[0]).todense()
            # eliminate slave dofs columns
            self.Tu = np.delete(self.Tu, self.find_dofs(self.dofs_s), axis=1)
            # eliminate slave dofs from dofs
            self.dofs = np.delete(self.dofs, self.find_dofs(self.dofs_s), axis=0)
            # update element incidence matrices based on master-slave equations
            for e, element in enumerate(self.my_elements, start=1):
                if e % 100 == 0:
                    print(f'update Zu element {e} of {len(self.my_elements)}')
                element.Zu = coo_array(element.Zu @ self.Tu, dtype=np.int8)

        # number of dofs
        self.ndof = self.dofs.shape[0]

        # compute the mass, stiffness and damping matrix
        self.compute_MCK_matricies()

        # compute the proportional damping matrix
        self.compute_Cp()

        # compute collocation matrix for controlled forces
        self.compute_Bf()

        # compute collocation matrix for controlled displacements
        self.compute_Bu()

        # initialization of state vectors
        self.u = np.zeros((self.ndof), dtype=int)
        self.v = self.u.copy()
        self.a = self.u.copy()
        self.r = self.u.copy()
        self.f = self.u.copy()
        self.l = np.zeros((self.dofs_u.shape[0]), dtype=int)

        # Extended list of model DOFs for plot (6 DOFs per node)
        self.dofs_x = np.column_stack((np.repeat(self.my_nodes.nodal_labels, 6),
                                       np.tile(np.array([1, 2, 3, 4, 5, 6]), 
                                               self.my_nodes.nodal_labels.shape[0]))).astype(int)

        # indices of the model dofs on the extended dofs
        self.ind_x = self.find_dofs_x(self.dofs)

        # number of extended dofs
        self.ndofs_x = self.dofs_x.shape[0]
        
        # extended quantities for mesh plots
        self.u_x = np.zeros((self.ndofs_x), dtype=int)
        self.v_x = self.u_x.copy()
        self.a_x = self.u_x.copy()

        # check the linearity of the model
        self.linear_M = True
        self.linear_C = True
        self.linear_K = True
    
        for my_element in self.my_elements:
            self.linear_K = self.linear_K & my_element.linear_K
            self.linear_M = self.linear_M & my_element.linear_M
            self.linear_C = self.linear_C & my_element.linear_C

        # if the model is linear in M, C and K it is fully linear
        self.linear = self.linear_K & self.linear_M & self.linear_C


#%% concurrent processes 
    def compute_collocations(self):
        with concurrent.futures.ThreadPoolExecutor() as executor:
                executor.map(self.proces_collocations, self.my_elements)
        
        # global cache_collocations

        # Zu = cache_collocations.get("Zu", None)
        # Zq = cache_collocations.get("Zq", None)

        # if Zu is None or Zq is None:
        #     print('No cache')
        #     with concurrent.futures.ThreadPoolExecutor() as executor:
        #         executor.map(self.proces_collocations, self.my_elements)

        #     cache_collocations["Zu"] = [elem.Zu for elem in self.my_elements]
        #     cache_collocations["Zq"] = [elem.Zq for elem in self.my_elements]

        # else:
        #     print('use cache')
        #     for elem, zu_val, zq_val in zip(self.my_elements, Zu, Zq):
        #         elem.Zu = zu_val
        #         elem.Zq = zq_val

    def proces_collocations(self, element):
        element.compute_Zu(self.dofs)
        element.compute_Zq(self.dofs_q)

    def compute_MCK_matricies(self):
        dofs0   = np.zeros_like(self.dofs[:, 0])    # Assumes dofs is (N, D), this gives length-N
        dofs0_q = np.zeros_like(self.dofs_q[:, 0])

        args = (dofs0, dofs0, dofs0_q, 0.0, 0)

        with concurrent.futures.ThreadPoolExecutor() as executor:
            futures = [executor.submit(self.compute_M, *args),
                       executor.submit(self.compute_K, *args),
                       executor.submit(self.compute_C, *args),
                       ]
            concurrent.futures.wait(futures)


#%% this method extract the parameters and assigns default values
    def extract_pars(self,pars):

        # storage of input for packing/unpacking
        self.my_pars = pars

        self.step = int(pars.get('step', 10)) # number of simulation steps
        self.dt = float(pars.get('dt', 1.0)) # time step size

        self.dofs_c = pars.get('dofs_c', np.zeros((0, 2), dtype=np.int32)).astype(np.int32) # list of constrained dofs
        self.dofs_m = pars.get('dofs_m', np.zeros((0, 2), dtype=np.int32)).astype(np.int32) # list of master dofs
        self.dofs_s = pars.get('dofs_s', np.zeros((0, 2), dtype=np.int32)).astype(np.int32) # list of slave dofs
        self.dofs_f = np.array(pars.get('dofs_f', np.zeros((0, 2), dtype=np.int32)), dtype=np.int32)  # list of force-controlled dofs
        self.dofs_u = np.array(pars.get('dofs_u', np.zeros((0, 2), dtype=np.int32)), dtype=np.int32)  # list of displacement-controlled dofs
        self.dofs_q = np.array(pars.get('dofs_q', np.zeros((0, 2), dtype=np.int32)), dtype=np.int32)  # list of temperature-controlled dofs

         # Determine the number of steps from g_f, g_u, and g_q
        steps = []
        if 'g_f' in pars: steps.append(np.shape(pars['g_f'])[-1]) 
        if 'g_u' in pars: steps.append(np.shape(pars['g_u'])[-1]) 
        if 'g_q' in pars: steps.append(np.shape(pars['g_q'])[-1]) 

        # Set the number of steps, defaulting to 10 if no steps are provided
        self.step = max(steps) if steps else int(pars.get('step', 10))

        # Handle time step size
        self.dt = float(pars.get('dt', 1.0))

        # loading/force/temperature history applied to force/displacement/temperature-controlled dofs
        self.g_f = np.array(pars.get('g_f', np.zeros((self.dofs_f.shape[0], self.step), dtype=np.float64)), dtype=np.float64)
        self.g_u = np.array(pars.get('g_u', np.zeros((self.dofs_u.shape[0], self.step), dtype=np.float64)), dtype=np.float64)
        self.g_q = np.array(pars.get('g_q', np.zeros((self.dofs_q.shape[0], self.step), dtype=np.float64)), dtype=np.float64)

        # Ensure that g_f, g_u, and g_q are 2D arrays
        if len(self.g_f.shape) == 1: self.g_f = np.array([self.g_f])
        if len(self.g_u.shape) == 1: self.g_u = np.array([self.g_u])
        if len(self.g_q.shape) == 1: self.g_q = np.array([self.g_q])

        self.damping_model = pars.get('damping_model', 'none') # damping model - 'none' is proportional
        self.alpha = float(pars.get('alpha', 0.0)) # mass-proportional damping coefficient
        self.beta = float(pars.get('beta', 0.0)) # stiffness-proportional damping coefficient

        '''
        # damping ratio for modal damping
        if 'zeta' in pars:
            self.zeta = pars['zeta']
        else:
            self.zeta = 0.0
        '''
        # checks on the input
        if self.dofs_s.shape[0] != self.dofs_m.shape[0]:
            raise Exception('number of master/slave dofs do not match!')

#%% compute the model internal force
    def compute_f(self,u,v,a,q,t,i):
        
        f = np.zeros((self.dofs.shape[0]))

        # loop over the elements
        for my_element in self.my_elements:

            # compute the element restoring force vector
            fe = my_element.compute_f(my_element.Zu @ u, my_element.Zu @ v, my_element.Zu @ a, my_element.Zq @ q, t, i)
            # assemble the model restoring force vector
            f += my_element.Zu.transpose() @ fe

        # return the model restoring force vector
        return f            


#%% compute the model restoring force
    def compute_r(self,u,v,q,t,i):

        # initialize the restoring force vector
        r = np.zeros((self.dofs.shape[0]))

        # loop over the elements
        for my_element in self.my_elements:

            # compute the element restoring force vector
            re = my_element.compute_r(my_element.Zu @ u,my_element.Zu @ v, my_element.Zq @ q, t, i)

            # assemble the model restoring force vector
            r += my_element.Zu.transpose() @ re

        # return the model restoring force vector
        return r

#%% compute the model stiff and damping matrix
    def __compute_matrix(self, matrix_type, u, v, q, t, i):
        # Preallocate lists for assembling the matrix
        data = [[]] * len(self.my_elements)
        row_indices = data.copy()
        col_indices = data.copy()

        # Loop over the elements
        for idx, my_element in enumerate(self.my_elements):
            # Compute local quantities
            u_local = my_element.Zu @ u
            v_local = my_element.Zu @ v
            q_local = my_element.Zq @ q

            # Compute the local matrix using the matrix_type
            local_matrix = csr_array(matrix_type(my_element, u_local, v_local, q_local, t, i))

            # Assemble the global matrix
            coo_assembled = coo_array(my_element.Zu.transpose() @ local_matrix @ my_element.Zu)

            # Append data, rows, and columns to the respective lists
            data[idx]        = coo_assembled.data
            row_indices[idx] = coo_assembled.row
            col_indices[idx] = coo_assembled.col

        # Concatenate the collected data and indices
        data = np.concatenate(data)
        row_indices = np.concatenate(row_indices)
        col_indices = np.concatenate(col_indices)

        # Construct the final matrix in COO format, then convert to CSR
        return coo_array((data, (row_indices, col_indices)), shape=(len(self.dofs), len(self.dofs))).tocsr()
       
    def compute_K(self, u, v, q, t, i):
        # Compute the stiffness matrix
        if self.linear_K == False:
            self.K = self.__compute_matrix(core_elem.compute_K, u, v, q, t, i)
        return self.K
    
    def compute_C(self, u, v, q, t, i):
        # Compute the damping matrix
        if self.linear_C == False:
            self.C = self.__compute_matrix(core_elem.compute_C, u, v, q, t, i)
        return self.C

    def compute_M(self, u, v, q, t, i):
        # Compute the mass matrix
        if self.linear_M == False:
            self.M = self.__compute_matrix(core_elem.compute_M, u, v, q, t, i)
        return self.M

# #%% compute the model mass matrix
#     def compute_M(self):
#         # Preallocate lists for assembling the mass matrix
#         data = [[]] * len(self.my_elements)
#         row_indices =  data.copy()
#         col_indices =  data.copy()

#         # Loop over the elements
#         for idx, my_element in enumerate(self.my_elements):
#             # Compute element mass matrix
#             Me = csr_array(my_element.compute_M())

#             # Collect the non-zero values and indices for final assembly
#             coo_assembled_M = coo_array(my_element.Zu.transpose() @ Me @ my_element.Zu)

#             # Append data, rows, and columns to the respective lists
#             data[idx]        = coo_assembled_M.data
#             row_indices[idx] = coo_assembled_M.row
#             col_indices[idx] = coo_assembled_M.col

#         # Concatenate the collected data and indices
#         data = np.concatenate(data)
#         row_indices = np.concatenate(row_indices)
#         col_indices = np.concatenate(col_indices)

#         # Construct the final mass matrix in COO format, then convert to CSR
#         self.M = coo_array((data, (row_indices, col_indices)), shape=(len(self.dofs), len(self.dofs))).tocsr()

#         return self.M

#%% compute the proportional damping
    def compute_Cp(self):

        if self.damping_model == 'none':
            self.Cp = coo_array((len(self.dofs),len(self.dofs))).tocsr()

        elif self.damping_model == 'proportional':
            
            # compute the mass and stiffness matrices
            if np.shape(self.M) == 0 and np.shape(self.K) == 0:

                # initialize the displacement, velocity and acceleration
                u = np.zeros((self.dofs.shape[0],1))
                v = u.copy()
                q = np.zeros((self.dofs_q.shape[0],1))

                self.M = self.compute_M(u,v,q,0.0,0)
                self.K = self.compute_K(u,v,q,0.0,0)

            # compute the proportional damping
            self.Cp = self.M*self.alpha + self.K*self.beta     

        # return the proportional damping matrix
        return self.Cp    

#%% compute the collocation matrix for controlled forces or controlled displacements
    def __compute_collocation_B(self, ind_dofs, ind_dofs_uf):

        # Convert to numpy arrays for broadcasting, if not already
        dofs    = np.array(ind_dofs)
        dofs_uf = np.array(ind_dofs_uf)

        # Create a boolean matrix where True represents a match between dofs and dofs_u/dofs_f
        matches = np.all(dofs[:, None] == dofs_uf, axis=2)

        # Find indices where matches are True
        rows, cols = np.where(matches)

        # Create and return the sparse matrix
        return coo_array((np.ones_like(rows, dtype=bool), (rows, cols)), shape=(dofs.shape[0], dofs_uf.shape[0])).tocsc()

    def compute_Bf(self):
        self.Bf = self.__compute_collocation_B(self.dofs, self.dofs_f)

    def compute_Bu(self):
        self.Bu = self.__compute_collocation_B(self.dofs, self.dofs_u)

#%% reset the model
    def reset(self):
        
        self.u = np.zeros_like(self.u)
        self.v = np.zeros_like(self.v)
        self.a = np.zeros_like(self.a)
        self.r = np.zeros_like(self.r)
        self.f = np.zeros_like(self.f)
        self.l = np.zeros_like(self.l)

        self.u_x = np.zeros_like(self.u_x)
        self.v_x = np.zeros_like(self.v_x)
        self.a_x = np.zeros_like(self.a_x)

        # loop over the elements
        for my_element in self.my_elements:

            # reset the element
            my_element.reset()

#%% compute the modal analysis of the model
    def compute_modal(self,mode=1):

        # compute the mass and stiffness matrices
        if np.shape(self.M) == 0 and np.shape(self.K) == 0:

            # initialize the displacement, velocity and acceleration
            u = np.zeros((self.dofs.shape[0]))
            v = u.copy()
            q = np.zeros((self.dofs_q.shape[0]))
                
            # compute the mass and stiffness matrices
            self.M = self.compute_M(u,v,q,0.0,0)
            self.K = self.compute_K(u,v,q,0.0,0)
        
        global cache_eigenmodes
        v0  = cache_eigenmodes.get("phi", None)
        
        # solve the eigenvalue problem
        self.omega, self.phi = spla.eigsh(self.K, M=self.M, k=mode, which='LM', sigma=0, v0=v0, mode='normal')

        cache_eigenmodes["phi"] = self.phi[:, 0]
        
        # extended modal shape (for plots)
        self.phi_x = np.zeros((self.dofs_x.shape[0],self.phi.shape[1]))
        self.phi_x[self.ind_x,:] = self.phi

        # Compute natural frequencies (in Hz)
        self.omega = np.real(np.sqrt(self.omega))/(2*np.pi)

        return self.omega, self.phi
    
#%% compute the modal stress/strain
    # def compute_modal_se(self,mode=1):

    #     # compute the mass and stiffness matrices
    #     if not hasattr(self, "omega") and not hasattr(self, "phi"):
    #         self.omega, self.phi = self.compute_modal(mode)

    #     for i in range(len(self.my_elements)):
    #         if hasattr(self.my_elements[i], 'B'):
    #             self.my_elements[i].e_phi = []
    #             self.my_elements[i].s_phi = []
    #             for j in range(np.shape(self.phi)[1]):
    #                 B   = self.my_elements[i].B[:,self.find_dofs_e(self.my_elements[i].dofs)]
    #                 phi = self.phi[self.find_dofs(self.my_elements[i].dofs), j]

    #                 self.my_elements[i].e_phi.append(B @ phi)
    #                 self.my_elements[i].s_phi.append(self.my_elements[i].D @ self.my_elements[i].e_phi[j].T)

    #     return self.omega, self.phi
    
    def compute_modal_ss(self,mode_sel=1,dofs_sel=np.array([[1,1]])):
    
        # Convert to a numpy array
        mode_sel = np.atleast_1d(mode_sel)

        if np.shape(dofs_sel)[0] < np.shape(mode_sel)[0]:
            raise ValueError("Number of selected dofs (" + str(np.shape(dofs_sel)[0]) + 
                            ") and selected modes (" + str(np.shape(mode_sel)[0]) + ") have to match"
                            )

        # compute the mass and stiffness matrices
        if not hasattr(self, "omega") and not hasattr(self, "phi"):
            self.omega, self.phi = self.compute_modal(mode_sel.max())

        # Indicies of the selected dofs
        idxs_sel = self.find_dofs(dofs_sel)

        # Select the dofs
        if np.shape(self.phi)[1] == 1:
            phi_m = np.atleast_2d(self.phi[idxs_sel,:])
        else:   
            phi_m = self.phi[idxs_sel,:]

        # Select the modes
        phi_m = phi_m[:,mode_sel-1]

        # modal shapes for virtual sensing
        self.phi_vs = self.phi[:,mode_sel-1] @ np.linalg.solve(phi_m.T @ phi_m, phi_m.T)

        # Loop through all elements
        for i in range(len(self.my_elements)):
            if hasattr(self.my_elements[i], 'B'):
                self.my_elements[i].e_phi = self.my_elements[i].B @ self.my_elements[i].Zu @ self.phi_vs
                self.my_elements[i].s_phi = self.my_elements[i].D @ self.my_elements[i].e_phi

        return self.omega, self.phi
    
    # def compute_ss_modes(self,dofs_sel,mode_sel):

    #     # if mode_sel > dofs_sel trow an error

    #     self.compute_modal(mode_sel.max())
    #     idxs_sel = self.find_dofs(dofs_sel)

    #     print(self.phi)
    #     print(idxs_sel)
    #     print(mode_sel)

    #     phi_m = self.phi[idxs_sel,mode_sel-1] # modal shape on measured dofs
    #     self.phi_vs = self.phi[:,mode_sel-1] @ np.linalg.inv((phi_m.T @ phi_m)) @ phi_m.T # modal shapes for virtual sensing

    #     print(phi_m)
    #     print(self.phi_vs)

    #     for i in range(len(self.my_elements)):
    #         if hasattr(self.my_elements[i], 'B'):
    #             #self.my_elements[i].e_phi = np.zeros((self.my_elements[i].B.shape[0],dofs_sel))
    #             #self.my_elements[i].s_phi = np.zeros((self.my_elements[i].D.shape[0],dofs_sel))
    #             #for j in range(np.shape(phi_vs)[1]):
    #             self.my_elements[i].e_phi = self.my_elements[i].B @ self.my_elements[i].Zu @ self.phi_vs
    #             self.my_elements[i].s_phi = self.my_elements[i].D @ self.my_elements[i].e_phi
                
    #             #phi = self.phi[self.find_dofs(self.my_elements[i].dofs), j]
    #             #self.my_elements[i].e_phi.append(B @ phi)
    #             #self.my_elements[i].s_phi.append(self.my_elements[i].D @ self.my_elements[i].e_phi[j].T)

    #     return self.omega, self.phi
    
#%% extract the dof indices of selected dofs
    @jax.jit
    def __compute_find_dofs(dofs, dofs_sel):
        # Force "dofs_sel" to a 2D array
        dofs_sel = jnp.atleast_2d(dofs_sel)

        # Convert to JAX arrays
        dofs     = jnp.array(dofs)
        dofs_sel = jnp.array(dofs_sel)

        def find_index(dofs_sel):
            match = jnp.all(dofs == dofs_sel, axis=1)
            return jnp.argmax(match),  match

        # Use jax.vmap to apply the helper function over all selected DOFs
        idx, found = jax.vmap(find_index)(dofs_sel)
        rows = jnp.any(found, axis=1)

        return jnp.array(idx), rows, rows.all()

    def __check_find_dofs(self, dofs, dofs_sel):
        idx,  rows, all_true = model.__compute_find_dofs(dofs, dofs_sel)
        if not all_true: idx = idx[rows]
        return np.array(idx), rows
    
    # Find dofs in the constrained system
    def find_dofs(self, dofs_sel):
        return self.__check_find_dofs(self.dofs, dofs_sel)[0]
    
    # Find dofs in the full plotting system
    def find_dofs_x(self, dofs_sel):
        return self.__check_find_dofs(self.dofs_x, dofs_sel)[0]

    # Find indicies of dofs present/not present in the system
    def find_dofs_e(self, dofs_sel):
        return self.__check_find_dofs(self.dofs, dofs_sel)[1]


#%% extract the dof indices of selected nodes
    def find_nodes(self, nodes_sel):

        # force "nodes_sel" to a 2d array
        nodes_sel = np.atleast_2d(nodes_sel)

        nodes_sel = np.array(nodes_sel)
        
        nodes_in_dofs = np.array(self.dofs)[:, 0]
        matches = np.isin(nodes_in_dofs, nodes_sel)
        
        # Get the indices of matching elements
        idxs_sel = np.atleast_2d(np.where(matches)[0])

        if np.min(np.shape(nodes_sel)) != np.shape(idxs_sel)[0]:
            raise Exception("Warning! Number of selected nodes, do not match with detected indices")
        
        return idxs_sel

#%% extract the dof indices of selected directions
    def find_dirs(self,dirs_sel):
        dirs_sel = np.array(dirs_sel)
        
        nodes_in_dofs = np.array(self.dofs)[:, 1]
        matches = np.isin(nodes_in_dofs, dirs_sel)
        
        # Get the indices of matching elements
        idxs_sel = np.where(matches)[0]
        
        return idxs_sel
        
#%% post-processing
    def dump_to_paraview(self):
        # save nodal coordinates
        self.my_nodes.dump_to_paraview()
        # save element results
        for element in self.my_elements:
            element.dump_to_paraview()
        
    # dump the model results to the elements
    def dump_to_elements(self):
        for element in self.my_elements:
            element.u = element.Zu @ self.u
            element.v = element.Zu @ self.v
            element.q = element.Zq @ self.q
            # compute the element results (e.g., stress/strain)
            element.compute_results()
            
    #%% plots
    def plot(self, labels=False, rotate=(30,30), figsize=8, zoom=1.0, axis='on', response=None, scale=1):
        # plotting nodes
        ax = self.my_nodes.plot(labels=labels, rotate=rotate, figsize=figsize, zoom=zoom, axis=axis)
        
        # Collect all lines and surfaces for Line3DCollection and Poly3DCollection
        all_lines, all_surfaces = [], []
        for plot_properties in (element.plot(ax) for element in self.my_elements):
            
            # Applying line and surface features
            if plot_properties is None: continue  # Skip elements that return None (MCK elements)
            lines, surfaces = plot_properties
            if lines is not None:    all_lines.extend(lines)
            if surfaces is not None: all_surfaces.extend(surfaces)

        # Plotting response, if response is provided
        all_response_lines, all_response_surfaces = [], []
        if response is not None:
            # modal response expanded to full dofs array
            res = np.zeros([np.shape(self.dofs_x)[0]])
            res[self.find_dofs_x(self.dofs)] = np.atleast_1d(response)

            def process_element(element):
                idx = self.find_dofs_x(element.dofs)

                # Maximum dof number app
                max_dof_num = max(element.dofs[:, 1])

                # Storing model responses
                res_x = res[idx[0::max_dof_num]] * scale + np.atleast_1d(element.nodal_coords[:, 0])
                res_y = res[idx[1::max_dof_num]] * scale + np.atleast_1d(element.nodal_coords[:, 1])
                res_z = res[idx[2::max_dof_num]] * scale + np.atleast_1d(element.nodal_coords[:, 2])

                # Extracting element plot properties
                plot_properties = element.plot(ax, x=res_x, y=res_y, z=res_z, color='r')

                # Return the plot properties
                return plot_properties

            # Use ThreadPoolExecutor for concurrency
            with concurrent.futures.ThreadPoolExecutor() as executor:
                futures = [executor.submit(process_element, element) for element in self.my_elements]

                for future in futures:
                    plot_properties = future.result()

                    # Applying line and surface features
                    if plot_properties is None:
                        continue  # Skip elements that return None (MCK elements)
                    lines, surfaces = plot_properties
                    if lines: all_response_lines.extend(lines)
                    if surfaces: all_response_surfaces.extend(surfaces)

        # Combine all_surfaces and all_response_surfaces
        combined_lines = all_lines + all_response_lines
        combined_surfaces = all_surfaces + all_response_surfaces

        # Add all collected lines and surfaces to collections for plotting
        if combined_surfaces:
            facecolors = ['cyan'] * len(all_surfaces) + ['red'] * len(all_response_surfaces)
            poly_collection = Poly3DCollection(combined_surfaces, facecolors=facecolors, shade=True, lightsource=None, alpha=0.5, edgecolor=None)
            ax.add_collection3d(poly_collection)
        if combined_lines:
            colors = ['black'] * len(all_lines) + ['red'] * len(all_response_lines)
            line_collection = Line3DCollection(combined_lines, color=colors, linewidth=0.5)
            ax.add_collection(line_collection)

        return ax



import importlib
from yafem.nodes import nodes
from yafem.model import model
from yafem.simulation import simulation

__all__ = ['nodes',
           'model',
           'simulation',
           'json',
           ]


#%% Import submodules

elem = importlib.import_module('.elem', __package__)




import numpy as np
import scipy as sp
import matplotlib.pyplot as plt
import matplotlib.patches as patches
from scipy.linalg import block_diag
from yafem.nodes import nodes
from yafem.elem.core_elem import core_elem
from yafem.elem.beam2d_gen_func import *

#%% element_beam2d class
class beam2d(core_elem):

    # class constructor
    def __init__(self, my_nodes, pars):

        # superclass constructor
        super().__init__(my_nodes,pars)

        self.linear_M = True
        self.linear_K = True
        self.linear_C = True

        # extract parameters and assign default values
        self.extract_pars(pars)

        # element dofs
        self.dofs = np.array([[self.nodal_labels[0],1],   
                              [self.nodal_labels[0],2],   
                              [self.nodal_labels[0],3],   
                              [self.nodal_labels[1],1],
                              [self.nodal_labels[1],2],
                              [self.nodal_labels[1],3]],dtype=np.int32)
        
        # # rotation matrix in the xy plane
        r = (self.nodal_coords[1,:2] - self.nodal_coords[0,:2])/self.L
        s = np.array([-r[1],r[0]])

        # Local reference system
        self.T = np.array([r, s])

        # global to local transformation matrix in the xy plane
        self.G = np.zeros((6, 6))
        self.G[0:2, 0:2] = self.T
        self.G[2:3, 2:3] = 1
        self.G[3:5, 3:5] = self.T
        self.G[5:6, 5:6] = 1

        # Variables
        variables = [self.A, 
                     self.E, 
                     self.I, 
                     self.L, 
                     self.alpha, 
                     self.fa, 
                     self.fb, 
                     self.k0a, 
                     self.k0b, 
                     self.rho, 
                     self.theta]

        self.Kl  = beam2d_gen_Kl(*variables)
        self.Ml  = beam2d_gen_Ml(*variables)
        self.rl  = beam2d_gen_rl(*variables)
        self.Nl  = beam2d_gen_Nl(*variables)
        self.Bl  = beam2d_gen_Bl(*variables)

        # cross-section stiffness matrix
        self.D = beam2d_gen_Dcs_mid(*variables)

        #%% Global reference

        # Stiffness matrix in global coordinate system
        self.K = self.G.T @ self.Kl @ self.G
        self.M = self.G.T @ self.Ml @ self.G

        # damping matrix in global coordinates
        self.C = np.zeros_like(self.K)

        # strain interpolation matrix in global coordinates
        self.B = self.Bl @ self.G

        # displacement interpolation matrix in global coordinates
        self.N = self.Nl @ self.G

        # local to global coordinate transformation
        self.r = self.G.T @ self.rl     
 
        #%% extract parameters
    def extract_pars(self, pars):
        
        # this is the element class used in packing/unpacking
        self.my_pars['elem'] = 'beam2d'

        self.A     = pars.get("A", 200.0) # Cross-sectional areal
        self.I     = pars.get("I", 1000.0) # second area moment
        self.E     = pars.get("E", 210e3) # Youngs modulus
        self.G1    = pars.get("G", 81e3) # Shear modulus
        self.k0a   = pars.get("k0a", 0.0) # Axial Winkler stiffness
        self.k0b   = pars.get("k0b", 0.0) # Lateral Winkler stiffness
        self.rho   = pars.get("rho", 7850/1e9) # Density
        self.fa    = pars.get("fa", 0.0) # Axial element destributed force
        self.fb    = pars.get("fb", 0.0) # Lateral element destributed force in y-direction
        self.alpha = pars.get("alpha", 0.0) # coefficient of thermal expansion 
        self.theta = pars.get("theta", 0.0) # Thermal loading
        self.nodal_labels = pars.get("nodal_labels", [1, 2])
        
        # extract nodal coordinates
        self.nodal_coords = self.my_nodes.find_coords(self.nodal_labels)
        self.L = np.linalg.norm(self.nodal_coords[1] - self.nodal_coords[0])
    
        # temperature controlled dofs
        self.dofs_q = np.array(pars.get("dofs_q", []), dtype=np.int32).reshape(-1, 2) if "dofs_q" in pars else np.zeros((0, 2), dtype=np.int32)


    #%% plot the element       
    def plot(self, ax, x=None, y=None, z=None, color='k-'):
        if x is None: x = self.nodal_coords[:, 0]
        if y is None: y = self.nodal_coords[:, 1]
        if z is None: z = self.nodal_coords[:, 2]

        # Collect lines
        lines = [[[x[0], y[0], 0],[x[1], y[1], 0]]]

        return lines, None



import numpy as np
import scipy as sp
import matplotlib.pyplot as plt
import matplotlib.patches as patches

class nodes:
#%% object constructor
    def __init__(self, pars):

        # extract parameters and assign default values
        self.extract_pars(pars)

        if self.nodal_coords.shape[0] == 1:
            self.model_size = 1
        else:    
            self.model_size = np.max(sp.spatial.distance.pdist(self.nodal_coords, 'euclidean'))
        
        # compute model center
        self.model_center = np.mean(self.nodal_coords,axis=0)

#%% extract nodal coordinates
    def find_coords(self, nodal_labels_sel):
        nodal_labels_sel = np.array(nodal_labels_sel).flatten()
        
        # Get the indices of each label in nodal_labels_sel
        indices = np.argsort(self.nodal_labels)
        sorted_labels = np.searchsorted(self.nodal_labels[indices], nodal_labels_sel)
        row_indices = indices[sorted_labels]

        # Extract the corresponding coordinates
        nodal_coords_subset = self.nodal_coords[row_indices]
        
        return nodal_coords_subset

#%% extract parameters and assign default values
    def extract_pars(self,pars):
        
        # storage of input for packing/unpacking
        self.my_pars = pars

        if 'nodal_data' in pars:

            # purging repeated nodal data
            pars['nodal_data'] = np.unique(pars['nodal_data'], axis=0)

            # checking for inconsistency in repeated nodal data
            check_labels = np.unique(pars['nodal_data'][:,0], axis=0,return_index=True,return_counts=True)
            check_coords = np.unique(pars['nodal_data'][:,1:], axis=0,return_index=True,return_counts=True)

            # number of inconsistent repeated labels and coords
            repeated_labels = check_labels[1][check_labels[2]>1]
            repeated_coords = check_coords[1][check_coords[2]>1]

            if (len(repeated_labels) != 0) or (len(repeated_coords) != 0):

                # Obtaining labels with repeated inconsistencies
                cases = pars['nodal_data'][repeated_coords,1:]
                cases_set = set(map(tuple, cases)) # Conversion to set for faster loop in list comprehension
                idx_bool = [tuple(i) in cases_set for i in pars['nodal_data'][:,1:]]
                labels_for_repeated_coords = pars['nodal_data'][idx_bool,0]
                labels_for_repeated_labels = pars['nodal_data'][repeated_labels,0]

                # if either inconsistent repeated labels or repeated coords
                if (len(repeated_labels) != 0) != (len(repeated_coords) != 0):
                   
                    if (len(repeated_labels) != 0) == True:
                        raise Exception("Warning! Repeated nodal label(s): " + str(set(map(int, labels_for_repeated_labels))))
                    
                    else: # (len(repeated_coords) != 0) == True:
                        raise Exception("Warning! Repeated nodal coordinates at label: " + str(set(map(int, labels_for_repeated_coords))))
     
                else: # if both inconsistent repeated labels and coords
                    raise Exception("Warning! Inconsistent repeated nodal label(s): " + str(set(map(int, labels_for_repeated_labels))) + 
                                      " and repeated nodal coordinates at label(s): " + str(set(map(int, labels_for_repeated_coords))))

            # extract nodal labels into 1d array
            self.nodal_labels = pars['nodal_data'][:,0].squeeze().astype(np.int32)
            
            # extract nodal coordinates into 2d array
            self.nodal_coords = pars['nodal_data'][:,1:].astype(np.float64)
            
        else:
            self.nodal_labels = np.array([1],dtype=np.int32)
            self.nodal_coords = np.array([[0.0,0.0,0.0]],dtype=np.float64)

        # here we set the missing dimensions to zero so that we can plot the nodes in 3D
        if self.nodal_coords.shape[1] < 3:
            self.nodal_coords = np.hstack((self.nodal_coords,np.zeros((self.nodal_coords.shape[0],3-self.nodal_coords.shape[1]),dtype=np.float64)))

#%% plot the nodes
    def plot(self, labels=True, rotate=(30,30), figsize=8, zoom=1.0, axis='on'):

        # Extracting data from the object
        nodal_coords = self.nodal_coords
        nodal_labels = self.nodal_labels
        model_center = self.model_center
        model_size   = self.model_size

        # Calculate initial limits
        xlim = [model_center[0] - model_size, model_center[0] + model_size]
        ylim = [model_center[1] - model_size, model_center[1] + model_size]
        zlim = [model_center[2] - model_size, model_center[2] + model_size]

        # Apply zoom factor
        x_range = xlim[1] - xlim[0]
        y_range = ylim[1] - ylim[0]
        z_range = zlim[1] - zlim[0]

        # Apply zoom
        zoom_factor = 1/zoom
        x_range_zoomed = x_range * zoom_factor
        y_range_zoomed = y_range * zoom_factor
        z_range_zoomed = z_range * zoom_factor

        # Create a 3D plot
        fig = plt.figure(figsize=(figsize, figsize))  # Adjust the size as needed       
        ax  = fig.add_subplot(111, projection='3d')

        # Switch for axis
        plt.axis(axis)
        
        # Plotting nodes
        ax.scatter(nodal_coords[:, 0], nodal_coords[:, 1], nodal_coords[:, 2], c='k', marker='.', depthshade=False)

        # Center the axes at the model center with zoom applied
        ax.set_xlim([model_center[0] - x_range_zoomed * 0.5, model_center[0] + x_range_zoomed * 0.5])
        ax.set_ylim([model_center[1] - y_range_zoomed * 0.5, model_center[1] + y_range_zoomed * 0.5])
        ax.set_zlim([model_center[2] - z_range_zoomed * 0.5, model_center[2] + z_range_zoomed * 0.5])

        # Setting labels
        ax.set_xlabel('x')
        ax.set_ylabel('y')
        ax.set_zlabel('z')

        ax.view_init(elev=rotate[0], azim=rotate[1])  # Adjust these values as needed

         # Plot nodal numbers if labels are enabled
        if labels == True and type(labels) == bool or type(labels) == int or type(labels) == float:
            # Setting fontsize
            if labels == True and type(labels) == bool:
                fontsize = 12
            else:
                fontsize = labels
         
            for i in range(len(nodal_coords)):
                x, y, z = nodal_coords[i]
                label = nodal_labels[i]  # Adjust for 1D array
                ax.text(x, y, z,str(label), fontsize=fontsize, fontweight='bold')

                # if nodal_labels[i] >= 42 and nodal_labels[i] <= 46:
                #     x, y, z = nodal_coords[i]
                #     label = nodal_labels[i]  # Adjust for 1D array
                #     ax.plot(nodal_coords[i, 0], nodal_coords[i, 1], nodal_coords[i, 2], c='r', marker='.',markersize=20)
                #     ax.text(x, y, z,str(label), fontsize=fontsize, fontweight='bold')

        return ax
        
    def dump_to_paraview(self):
        pass



import numpy as np
from yafem.nodes import nodes
from yafem.elem.core_elem import core_elem

#%% element_MCK class
class MCK(core_elem):

    # class constructor
    def __init__(self, my_nodes, pars):

        # superclass constructor
        super().__init__(my_nodes,pars)

        self.linear_M = True
        self.linear_K = True
        self.linear_C = True

        # extract parameters and assign default values
        self.extract_pars(pars)

    #%% extract parameters and assign default values
    def extract_pars(self,pars):

        # this is the element class used in packing/unpacking
        self.my_pars['elem'] = 'MCK'

        # stiffness matrix
        if 'K' in pars:
            self.K = pars['K'].astype(np.float64)
        else:
            raise Exception('the stiffness parameter must be defined')

        self.M = pars.get('M', np.zeros(self.K.shape).astype(np.float64)) # mass matrix
        self.C = pars.get('C', np.zeros(self.K.shape).astype(np.float64)) # damping matrix

        # list of dofs
        self.dofs = pars.get('dofs', np.transpose([np.arange(0, self.K.shape[0], dtype=np.int32), 
                                     np.zeros(self.K.shape[0], dtype=np.int32)])).astype(np.int32)

        # temperature controlled dofs
        self.dofs_q = pars.get('dofs_q', np.zeros((0, 2), dtype=np.int32)).astype(np.int32)